"""""
模块作用：定义量子码抽象基类
"""""
import copy
from abc import ABC, abstractmethod
import numpy as np
from extendedstim.Physics.Operator import Operator


class QuantumCode(ABC):
    """""
    抽象基类，定义量子码的基本属性和方法
    self.generators：list对象，稳定子列表
    self.physical_number：int对象，物理二能级系统数目
    self.checker_number：int对象，校验子数目
    """""


    #%%  CHAPTER：====构造方法====
    def __init__(self,generators:list,physical_number:int):
        """""
        input.generators：稳定子列表
        input.physical_number：物理二能级系统数目
        """""

        ##  PART：检查输入合法性
        assert physical_number > 0
        physical_number=int(physical_number)
        assert all(isinstance(generator, Operator) for generator in generators)
        generators=[generator for generator in generators]

        ##  PART：赋值
        self.generators=generators
        self.physical_number=physical_number
        self.checker_number=len(self.generators)

    #%%  CHAPTER：====属性方法====
    ##  SECTION：----求校验矩阵----
    @property
    def check_matrix(self):
        """""
        output：GF(2)矩阵，按占据向量堆叠得到 (m,2n)
        """""
        matrix=Operator.get_matrix(self.generators, self.physical_number)
        return matrix

    ##  SECTION：----求校验矩阵的秩----
    @property
    def rank(self):
        """""
        output：int，rank(H)
        """""
        return np.linalg.matrix_rank(self.check_matrix)

    ##  SECTION：----求logical number----
    @property
    def logical_number(self):
        """""
        output：int = physical_number - rank
        """""
        return self.physical_number-self.rank

    ##  SECTION：----求码距----
    @property
    @abstractmethod
    def distance(self):
        """""
        output：int，最小非平凡逻辑算符权重
        """""
        pass

    ##  SECTION：----求逻辑算符----
    @property
    @abstractmethod
    def logical_operators(self):
        """""
        output：算符列表，成对或成组的独立逻辑算符
        """""
        pass

    #%%  CHAPTER：====对象方法====
    ##  SECTION：----修改索引----
    def index_map(self, index_map,physical_number):
        """""
        input.index_map：列表或数组，给出新索引映射
        output：无（原地修改生成元的索引）
        """""
        for generator in self.generators:
            generator.index_map(index_map)
        self.physical_number=physical_number

    ##  SECTION：----复制代码----
    def copy(self):
        """""
        output：深拷贝新实例
        """""
        return copy.deepcopy(self)


"""""
将一个量子码转换为测试线路，计算它的physical error rate与logical error rate之间的关系
"""""
from extendedstim.Code.BeyondCode.FermionicLatticeSurgery import FermionicLatticeSurgery
from extendedstim.Code.PrimitiveCode.MajoranaCode import MajoranaCode
from extendedstim.Circuit.Circuit import Circuit
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.PauliCSSCode import PauliCSSCode
from extendedstim.Code.PrimitiveCode.PauliCode import PauliCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator
from extendedstim.tools.TypingTools import isinteger
import numpy as np


#%%  CHAPTER：====将量子码转换为量子线路====
def Code2Circuit(code: MajoranaCode | PauliCode | FermionicLatticeSurgery, noise_model: str, cycle_number: int):
    """""
    code：要转换的量子码
    p_noise：去极化噪声发生的几率
    p_measure：测量噪声发生的几率
    noise_model：噪声模型，可选值为'phenomenological'（现象级噪声）或'circuit-level'（电路级噪声）
    cycle_number：循环次数
    """""

    ##  SECTION：----数据预处理----
    assert isinteger(cycle_number) and cycle_number >= 0
    assert isinstance(code, MajoranaCode) or isinstance(code, PauliCode) or isinstance(code, FermionicLatticeSurgery)

    ##  SECTION：----根据量子码类型选择不同的处理函数----
    ##  处理现象级噪声
    if noise_model == 'phenomenological':
        if isinstance(code, MajoranaCSSCode):
            return _MajoranaCSSCode2PhenomenologicalCircuit(code, cycle_number)
        elif isinstance(code, PauliCSSCode):
            return _PauliCSSCode2PhenomenologicalCircuit(code, cycle_number)
        elif isinstance(code, MajoranaCode):
            raise NotImplementedError
        elif isinstance(code, PauliCode):
            raise NotImplementedError
        else:
            raise NotImplementedError

    ##  处理电路级噪声
    elif noise_model == 'circuit-level':
        if isinstance(code, MajoranaCSSCode):
            return _MajoranaCSSCode2CircuitLevelCircuit(code, cycle_number)
        elif isinstance(code, PauliCSSCode):
            return _PauliCSSCode2CircuitLevelCircuit(code, cycle_number)
        elif isinstance(code, MajoranaCode):
            raise NotImplementedError
        elif isinstance(code, PauliCode):
            raise NotImplementedError
        elif isinstance(code, FermionicLatticeSurgery):
            return _FermionicLatticeSurgery2CircuitLevelCircuit(code, cycle_number)
        else:
            raise NotImplementedError

    ##  处理码容量
    elif noise_model == 'code-capacity':
        if isinstance(code, MajoranaCSSCode):
            return _MajoranaCSSCode2CodeCapacityCircuit(code)
        else:
            raise NotImplementedError

    ##  其他类型抛出异常
    else:
        raise ValueError('noise_model must be phenomenological, circuit-level, or code-capacity')


def _MajoranaCSSCode2CodeCapacityCircuit(code: MajoranaCSSCode) -> Circuit:
    """""
    input.code：一个MajoranaCSSCode
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    input.cycle_number：syndrome测量的次数
    """""

    ##  SECTION：----数据预处理----
    stabilizers_x = code.generators_x

    ##  这里我们检查N型稳定子，而不是Z型稳定子
    stabilizers_n = []
    for i in range(len(code.generators_z)):
        stabilizers_n.append(code.generators_x[i] @ code.generators_z[i])

    ##  我们检查逻辑N算符
    logical_x = code.logical_operators_x
    logical_z = code.logical_operators_z
    logical_occupy = [1j * logical_x[temp] @ logical_z[temp] for temp in range(len(logical_x))]  # 粒子数算符组作为逻辑算符组

    ##  计算数目
    majorana_number = code.physical_number  # fermionic sites的数目
    stabilizer_number = len(stabilizers_x) + len(stabilizers_n)  # 稳定子的数目

    ##  SECTION：----生成线路----
    ##  初始化
    circuit = Circuit()
    circuit.append({'name': 'TICK'})
    circuit.append({'name': 'TRAP', 'majorana_number': majorana_number, 'pauli_number': 0})
    circuit.append({'name': 'TICK'})

    ##  测量逻辑算符
    observable_include = []  # 记录可观测量的索引
    for i, logical_operator in enumerate(logical_occupy):
        circuit.append({"name": "MPP", "target": logical_operator})
        observable_include.append(len(circuit._measurements) - 1)

    ##  记录X型测量稳定子的值
    for i, stabilizer in enumerate(stabilizers_x):
        circuit.append({"name": "MPP", "target": stabilizer})

    ##  记录N型测量稳定子的值
    for i, stabilizer in enumerate(stabilizers_n):
        circuit.append({"name": "MPP", "target": stabilizer})

    ##  量子噪声信道
    for i in range(majorana_number):
        circuit.append({"name": "FDEPOLARIZE1", "target": i, "p": 0})

    ##  测量X型稳定子
    for i, stabilizer in enumerate(stabilizers_x):
        circuit.append({"name": "MPP", "target": stabilizer, 'p': 0})

    ##  测量Z型稳定子
    for i, stabilizer in enumerate(stabilizers_n):
        circuit.append({"name": "MPP", "target": stabilizer, 'p': 0})

    for i in range(stabilizer_number):
        circuit.append({"name": "DETECTOR", "target": [i - stabilizer_number, i - 2 * stabilizer_number]})

    ##  测量逻辑算符
    for i, logical_operator in enumerate(logical_occupy):
        circuit.append({"name": "MPP", "target": logical_operator})
        circuit.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit._measurements) - 1, observable_include[i]]})

    ##  SECTION：----返回线路----
    return circuit


#%%  CHAPTER：===将Majorana CSS code转换为现象级噪声下的测试线路===
def _MajoranaCSSCode2PhenomenologicalCircuit(code: MajoranaCSSCode, cycle_number: int) -> Circuit:
    """""
    input.code：一个MajoranaCSSCode
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    input.cycle_number：syndrome测量的次数
    """""

    ##  SECTION：----数据预处理----
    stabilizers_x = code.generators_x

    ##  这里我们检查N型稳定子，而不是Z型稳定子
    stabilizers_n = []
    for i in range(len(code.generators_z)):
        stabilizers_n.append(code.generators_x[i] @ code.generators_z[i])

    ##  我们检查逻辑N算符
    logical_x = code.logical_operators_x
    logical_z = code.logical_operators_z
    logical_occupy = [1j * logical_x[temp] @ logical_z[temp] for temp in range(len(logical_x))]  # 粒子数算符组作为逻辑算符组

    ##  计算数目
    majorana_number = code.physical_number  # fermionic sites的数目
    stabilizer_number = len(stabilizers_x) + len(stabilizers_n)  # 稳定子的数目

    ##  SECTION：----生成线路----
    ##  初始化
    circuit = Circuit()
    circuit.append({'name': 'TICK'})
    circuit.append({'name': 'TRAP', 'majorana_number': majorana_number, 'pauli_number': 0})
    circuit.append({'name': 'TICK'})

    ##  测量逻辑算符
    observable_include = []  # 记录可观测量的索引
    for i, logical_operator in enumerate(logical_occupy):
        circuit.append({"name": "MPP", "target": logical_operator})
        observable_include.append(len(circuit._measurements) - 1)

    ##  记录N型测量稳定子的值
    for i, stabilizer in enumerate(stabilizers_n):
        circuit.append({"name": "MPP", "target": stabilizer})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  量子噪声信道
        for i in range(majorana_number):
            circuit.append({"name": "FDEPOLARIZE1", "target": i, "p": 0})

        ##  测量X型稳定子
        for i, stabilizer in enumerate(stabilizers_x):
            circuit.append({"name": "MPP", "target": stabilizer, 'p': 0})

        ##  测量Z型稳定子
        for i, stabilizer in enumerate(stabilizers_n):
            circuit.append({"name": "MPP", "target": stabilizer, 'p': 0})

        ##  添加探测器，区分第一轮的情况
        if _ == 0:
            for i in range(stabilizer_number // 2):
                circuit.append({"name": "DETECTOR", "target": [i - stabilizer_number // 2, i - stabilizer_number - stabilizer_number // 2]})
        else:
            for i in range(stabilizer_number):
                circuit.append({"name": "DETECTOR", "target": [i - stabilizer_number, i - 2 * stabilizer_number]})

    ##  最后一轮测量稳定子横向测量
    for i in range(code.physical_number):
        circuit.append({"name": "MN", "target": i, 'p': 0})

    ##  添加N型测量稳定子的探测器
    for i, stabilizer in enumerate(stabilizers_n):
        target = []  # 确定检测的N测量与之前的测量
        stabilizer_now = MajoranaOperator([], [], 1)  # 记录实际测量乘积对应的算符
        for index in stabilizer.occupy_x:
            target.append(-code.physical_number + index)
            stabilizer_now = stabilizer_now @ MajoranaOperator([index], [index], 1j)
        target.append(-code.physical_number - stabilizer_number // 2 + i)  # 添加之前的N测量
        if stabilizer_now.coff != stabilizer.coff:
            target.append('negative')
        circuit.append({'name': 'DETECTOR', 'target': target})

    ##  测量逻辑算符
    for i, logical_operator in enumerate(logical_occupy):
        circuit.append({"name": "MPP", "target": logical_operator})
        circuit.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit._measurements) - 1, observable_include[i]]})

    ##  ----返回线路----
    return circuit


#%%  CHAPTER：====将Majorana CSS code转换为线路级噪声下的测试线路====
def _MajoranaCSSCode2CircuitLevelCircuit(code:MajoranaCSSCode, cycle_number:int)->Circuit:
    """""
    input.code：一个MajoranaCSSCode
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    input.cycle_number：syndrome测量的次数
    """""

    ##  SECTION：----数据预处理----
    ##  获取稳定子
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z
    stabilizers_n=[]
    for i in range(len(code.generators_z)):
        stabilizers_n.append(code.generators_x[i]@code.generators_z[i])

    ##  获取逻辑算符
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z
    logical_occupy=[1j*logical_x[temp]@logical_z[temp] for temp in range(len(logical_x))]  # 粒子数算符组作为逻辑算符组

    ##  获取数目
    majorana_number=code.physical_number
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    pauli_number=stabilizer_number

    ##  SECTION：----生成线路----
    ##  初始化
    circuit = Circuit()
    circuit.append({'name':'TRAP','majorana_number':majorana_number,'pauli_number':pauli_number})
    circuit.append({'name':'TICK'})

    ##  测量逻辑算符
    observable_include= []  # 记录可观测量的索引
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        observable_include.append(len(circuit._measurements)-1)

    ##  确定第一轮算符的值
    for i,stabilizer in enumerate(stabilizers_n):
        circuit.append({"name":"MPP","target":stabilizer})
    circuit.append({'name':'TICK'})

    ##  添加第一轮噪声
    for i in range(majorana_number):
        circuit.append({"name": "FDEPOLARIZE1", "target": i, "p":0})
    for i in range(pauli_number):
        circuit.append({"name": "DEPOLARIZE1", "target": i, "p": 0})
    circuit.append({'name':'TICK'})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):
        ##  添加稳定子测量
        for i,stabilizer in enumerate(stabilizers_x):
            sequence_temp=_syndrome_majorana_css_measurement_circuit(stabilizer, i, 'x')
            for temp in sequence_temp:
                circuit.append(temp)
            circuit.append({'name':'TICK'})

        for i,stabilizer in enumerate(stabilizers_z):
            sequence_temp=_syndrome_majorana_css_measurement_circuit(stabilizer, i + len(stabilizers_x), 'z')
            for temp in sequence_temp:
                circuit.append(temp)
            circuit.append({'name':'TICK'})

        ##  第一轮只添加N检测器
        if _==0:
            for i in range(stabilizer_number // 2):
                circuit.append({"name": "DETECTOR","target": [i - stabilizer_number ,i-stabilizer_number // 2, i - 3*stabilizer_number//2]})
        else:

            ##  添加X检测器
            for i in range(stabilizer_number//2):
                circuit.append({"name":"DETECTOR","target":[i-stabilizer_number, i - 2*stabilizer_number]})

            ##  添加N检测器
            for i in range(stabilizer_number // 2):
                circuit.append({"name": "DETECTOR","target": [i - stabilizer_number ,i-stabilizer_number//2, i - 3*stabilizer_number // 2,i-2*stabilizer_number]})

    ##  最后一轮测量稳定子横向测量
    for i in range(code.physical_number):
        circuit.append({"name":"MN","target":i,'p':0})

    for i,stabilizer in enumerate(stabilizers_n):
        target=[]
        stabilizer_now=MajoranaOperator([],[],1)
        for index in stabilizer.occupy_x:
            target.append(-code.physical_number+index)
            stabilizer_now=stabilizer_now@MajoranaOperator([index],[index],1j)
        target.append(-code.physical_number - stabilizer_number // 2 + i)
        target.append(-code.physical_number - stabilizer_number + i)
        if stabilizer_now.coff!=stabilizer.coff:
            target.append('negative')
        circuit.append({'name':'DETECTOR','target':target})

    ##  测量逻辑算符
    for i,logical_operator in enumerate(logical_occupy):
        circuit.append({"name":"MPP","target":logical_operator})
        circuit.append({"name":"OBSERVABLE_INCLUDE","target":[len(circuit._measurements)-1, observable_include[i]]})

    ##  SECTION：----返回线路----
    return circuit


#%%  CHAPTER：====将Fermionic Lattice Surgery code转换为线路级噪声下的测试线路====
def _FermionicLatticeSurgery2CircuitLevelCircuit(lattice_surgery: FermionicLatticeSurgery, cycle_number: int) -> Circuit:
    """""
    input.code：一个FermionicLatticeSurgeryCode
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    input.cycle_number：syndrome测量的次数
    """""

    ##  SECTION：----数据预处理----
    ##  获取稳定子
    irrelated_A_x = lattice_surgery.irrelated_stabilizers_A_x
    irrelated_A_z = lattice_surgery.irrelated_stabilizers_A_z
    irrelated_B_x = lattice_surgery.irrelated_stabilizers_B_x
    irrelated_B_z = lattice_surgery.irrelated_stabilizers_B_z
    related_A_z = lattice_surgery.related_stabilizers_A_z
    related_B_z = lattice_surgery.related_stabilizers_B_z
    modifiy_A_x = lattice_surgery.modify_stabilizers_A
    modifiy_B_x = lattice_surgery.modify_stabilizers_B
    gauge = lattice_surgery.gauge_stabilizers

    stabilizers_n = ([irrelated_A_x[temp] @ irrelated_A_z[temp] for temp in range(len(irrelated_A_x))] +
                     [irrelated_B_x[temp] @ irrelated_B_z[temp] for temp in range(len(irrelated_B_x))] +
                     [modifiy_A_x[temp] @ related_A_z[temp] for temp in range(len(modifiy_A_x))] +
                     [modifiy_B_x[temp] @ related_B_z[temp] for temp in range(len(modifiy_B_x))] +
                     gauge)
    stabilizers_z = irrelated_A_z + irrelated_B_z + related_A_z + related_B_z

    ##  获取逻辑算符
    logicals = lattice_surgery.bare_logical_operators
    logical_occupy = [
        MajoranaOperator.HermitianOperatorFromOccupy(np.concatenate((temp.occupy_x, temp.occupy_z)), np.concatenate((temp.occupy_x, temp.occupy_z)))
        for temp in logicals]

    ##  获取数目
    majorana_number = lattice_surgery.physical_number
    stabilizer_number_z = len(stabilizers_z)
    stabilizer_number_n = len(stabilizers_n)
    pauli_number = stabilizer_number_z + stabilizer_number_n

    ##  SECTION：----生成线路----
    ##  初始化
    circuit = Circuit()
    circuit.append({'name': 'TRAP', 'majorana_number': majorana_number, 'pauli_number': pauli_number})
    circuit.append({'name': 'TICK'})

    ##  测量逻辑算符
    observable_include = []  # 记录可观测量的索引
    for i, logical_operator in enumerate(logical_occupy):
        circuit.append({"name": "MPP", "target": logical_operator})
        observable_include.append(len(circuit._measurements) - 1)

    ##  添加第一轮噪声
    for i in range(majorana_number):
        circuit.append({"name": "FDEPOLARIZE1", "target": i, "p": 0})
    for i in range(pauli_number):
        circuit.append({"name": "DEPOLARIZE1", "target": i, "p": 0})
    circuit.append({'name': 'TICK'})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  添加稳定子测量
        for i, stabilizer in enumerate(stabilizers_z):
            sequence_temp = _syndrome_majorana_css_measurement_circuit(stabilizer, i, 'z')
            for temp in sequence_temp:
                circuit.append(temp)
            circuit.append({'name': 'TICK'})

        for i, stabilizer in enumerate(stabilizers_n):
            sequence_temp = _syndrome_majorana_css_measurement_circuit(stabilizer, i + len(stabilizers_z), 'n')
            for temp in sequence_temp:
                circuit.append(temp)
            circuit.append({'name': 'TICK'})

        ##  第一轮只添加N检测器
        if _ == 0:
            for i in range(stabilizer_number_n):
                circuit.append({"name": "DETECTOR", "target": [i - stabilizer_number_n]})
        else:

            ##  添加X检测器
            for i in range(stabilizer_number_z):
                circuit.append({"name": "DETECTOR",
                                "target": [i - stabilizer_number_n - stabilizer_number_z, i - 2 * (stabilizer_number_z + stabilizer_number_n)]})

            ##  添加N检测器
            for i in range(stabilizer_number_n):
                circuit.append({"name": "DETECTOR", "target": [i - stabilizer_number_n, i - 2 * stabilizer_number_n - stabilizer_number_z]})

    ##  最后一轮测量稳定子横向测量
    for i in range(lattice_surgery.physical_number):
        circuit.append({"name": "MN", "target": i, 'p': 0})

    for i, stabilizer in enumerate(stabilizers_n):
        target = []
        for index in stabilizer.occupy_x:
            target.append(-lattice_surgery.physical_number + index)
        target.append(-lattice_surgery.physical_number - stabilizer_number_n + i)
        circuit.append({'name': 'DETECTOR', 'target': target})

    ##  测量逻辑算符
    for i, logical_operator in enumerate(logical_occupy):
        circuit.append({"name": "MPP", "target": logical_operator})
        circuit.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit._measurements) - 1, observable_include[i]]})

    ##  SECTION：----返回线路----
    return circuit


#%%  CHAPTER：====将Pauli CSS code转换为现象级噪声下的测试线路====
def _PauliCSSCode2PhenomenologicalCircuit(code: PauliCSSCode, cycle_number: int) -> tuple[Circuit, Circuit]:
    """""
    input.code：一个PauliCSSCode
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    input.cycle_number：syndrome测量的次数
    """""

    ##  SECTION：----数据预处理----
    ##  获取稳定子
    stabilizers_x = code.generators_x
    stabilizers_z = code.generators_z

    ##  获取逻辑算符
    logical_x = code.logical_operators_x
    logical_z = code.logical_operators_z

    ##  获取数目
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    pauli_number = code.physical_number

    ##  SECTION：----生成线路----
    ##  强制初始化
    circuit_x = Circuit()
    circuit_z = Circuit()
    circuit_x.append({'name': 'TRAP', 'majorana_number': 0, 'pauli_number': pauli_number})
    circuit_x.append({'name': 'H', 'target': range(pauli_number)})
    circuit_z.append({'name': 'TRAP', 'majorana_number': 0, 'pauli_number': pauli_number})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  量子噪声信道
        for i in range(pauli_number):
            circuit_x.append({"name": "DEPOLARIZE1", "target": i, "p": 0})
            circuit_z.append({"name": "DEPOLARIZE1", "target": i, "p": 0})

        ##  测量稳定子
        for i, stabilizer in enumerate(stabilizers_x):
            circuit_x.append({"name": "MPP", "target": stabilizer, 'p': 0})
            circuit_z.append({"name": "MPP", "target": stabilizer, 'p': 0})
        for i, stabilizer in enumerate(stabilizers_z):
            circuit_z.append({"name": "MPP", "target": stabilizer, 'p': 0})
            circuit_x.append({"name": "MPP", "target": stabilizer, 'p': 0})

        ##  区分第一轮和其他轮
        if _ == 0:
            for i in range(stabilizer_number // 2):
                circuit_z.append({"name": "DETECTOR", "target": [-stabilizer_number // 2 + i]})
                circuit_x.append({"name": "DETECTOR", "target": [-stabilizer_number + i]})
        else:
            ##  添加检测器
            for i in range(stabilizer_number):
                circuit_z.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})
                circuit_x.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  最后一轮测量稳定子横向测量
    for i in range(code.physical_number):
        circuit_x.append({"name": "MPP", "target": PauliOperator([i], [], 1), 'p': 0})
        circuit_z.append({"name": "MPP", "target": PauliOperator([], [i], 1), 'p': 0})

    for i, stabilizer in enumerate(stabilizers_x):
        target = []
        for index in stabilizer.occupy_x:
            target.append(-code.physical_number + index)
        target.append(-code.physical_number - stabilizer_number + i)
        circuit_x.append({'name': 'DETECTOR', 'target': target})

    for i, stabilizer in enumerate(stabilizers_z):
        target = []
        for index in stabilizer.occupy_z:
            target.append(-code.physical_number + index)
        target.append(-code.physical_number - stabilizer_number // 2 + i)

        circuit_z.append({'name': 'DETECTOR', 'target': target})

    ##  测量逻辑算符
    for i in range(len(logical_x)):
        circuit_z.append({"name": "MPP", "target": logical_z[i]})
        circuit_x.append({"name": "MPP", "target": logical_x[i]})
        circuit_z.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_z._measurements) - 1]})
        circuit_x.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_x._measurements) - 1]})

    ##  SECTION：----返回线路----
    return circuit_x, circuit_z


#%%  CHAPTER：====将Pauli CSS code转换为电路级噪声下的测试线路====
def _PauliCSSCode2CircuitLevelCircuit(code: PauliCSSCode, cycle_number: int) -> tuple[Circuit, Circuit]:
    """""
    input.code：一个PauliCSSCode
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    input.cycle_number：syndrome测量的次数
    """""

    ##  SECTION：----数据预处理----
    ##  获取稳定子
    stabilizers_x = code.generators_x
    stabilizers_z = code.generators_z

    ##  获取逻辑算符
    logical_x = code.logical_operators_x
    logical_z = code.logical_operators_z

    ##  获取数目
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    data_number = code.physical_number
    pauli_number = data_number + stabilizer_number

    ##  SECTION：----生成线路----
    ##  强制初始化
    circuit_x = Circuit()
    circuit_z = Circuit()
    circuit_x.append({'name': 'TRAP', 'majorana_number': 0, 'pauli_number': pauli_number})
    circuit_z.append({'name': 'TRAP', 'majorana_number': 0, 'pauli_number': pauli_number})
    circuit_z.append({'name': 'TICK'})
    circuit_x.append({'name': 'TICK'})
    circuit_x.append({'name': 'H', 'target': list(range(data_number))})

    ##  施加第一轮噪声
    for i in range(pauli_number):
        circuit_z.append({"name": "DEPOLARIZE1", "target": i, "p": 0})
        circuit_x.append({"name": "DEPOLARIZE1", "target": i, "p": 0})

    circuit_z.append({'name': 'TICK'})
    circuit_x.append({'name': 'TICK'})

    ##  循环多轮，多轮测量错误与量子噪声信道
    for _ in range(cycle_number):

        ##  测量稳定子
        for i, stabilizer in enumerate(stabilizers_x):
            sequence_temp = _syndrome_pauli_css_measurement_circuit(stabilizer, i + data_number, 'x')
            for temp in sequence_temp:
                circuit_z.append(temp)
                circuit_x.append(temp)
            circuit_z.append({'name': 'TICK'})
            circuit_x.append({'name': 'TICK'})
        for i, stabilizer in enumerate(stabilizers_z):
            sequence_temp = _syndrome_pauli_css_measurement_circuit(stabilizer, i + data_number + len(stabilizers_x), 'z')
            for temp in sequence_temp:
                circuit_z.append(temp)
                circuit_x.append(temp)
            circuit_z.append({'name': 'TICK'})
            circuit_x.append({'name': 'TICK'})
        ##  添加检测器
        if _ == 0:
            for i in range(stabilizer_number // 2):
                circuit_z.append({"name": "DETECTOR", "target": [-stabilizer_number // 2 + i]})
                circuit_x.append({"name": "DETECTOR", "target": [-stabilizer_number + i]})
        else:
            for i in range(stabilizer_number):
                circuit_z.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})
                circuit_x.append({"name": "DETECTOR", "target": [-i - 1, -i - stabilizer_number - 1]})

    ##  最后一轮测量稳定子横向测量
    for i in range(code.physical_number):
        circuit_x.append({"name": "MPP", "target": PauliOperator([i], [], 1), 'p': 0})
        circuit_z.append({"name": "MPP", "target": PauliOperator([], [i], 1), 'p': 0})

    for i, stabilizer in enumerate(stabilizers_x):
        target = []
        for index in stabilizer.occupy_x:
            target.append(-code.physical_number + index)
        target.append(-code.physical_number - stabilizer_number + i)

        circuit_x.append({'name': 'DETECTOR', 'target': target})

    for i, stabilizer in enumerate(stabilizers_z):
        target = []
        for index in stabilizer.occupy_z:
            target.append(-code.physical_number + index)
        target.append(-code.physical_number - stabilizer_number // 2 + i)

        circuit_z.append({'name': 'DETECTOR', 'target': target})

    ##  测量逻辑算符
    for i in range(len(logical_x)):
        circuit_z.append({"name": "MPP", "target": logical_z[i]})
        circuit_x.append({"name": "MPP", "target": logical_x[i]})
        circuit_z.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_z._measurements) - 1]})
        circuit_x.append({"name": "OBSERVABLE_INCLUDE", "target": [len(circuit_x._measurements) - 1]})

    ##  SECTION：----返回线路----
    return circuit_x, circuit_z


# %%  CHAPTER：====生成Majorana CSS stabilizer测量线路====
def _syndrome_majorana_css_measurement_circuit(stabilizer: MajoranaOperator, qubit_index: int, type: str) -> list[dict]:
    """""
    input.stabilizer：一个MajoranaOperator，代表stabilizer
    input.qubit_index：一个整数，代表测量的qubit索引
    input.type：一个字符串，代表测量的类型，只能是'x'或'X'或'z'或'Z'
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    """""

    ##  SECTION：----数据预处理----
    sequence = []  # 线路序列
    flag = True  # 门类型标志

    ##  判断稳定子类型
    if type == 'x' or type == 'X':
        occupy = stabilizer.occupy_x
    elif type == 'z' or type == 'Z':
        occupy = stabilizer.occupy_z
    elif type == 'n' or type == 'N':
        assert np.all(stabilizer.occupy_z == stabilizer.occupy_x)
        occupy = stabilizer.occupy_x
    else:
        raise ValueError

    ##  SECTION：----生成线路----
    ##  对于N型的探测
    if type == 'n' or type == 'N':
        for j in range(len(occupy)):
            majorana_index_now = occupy[j]
            sequence.append({'name': 'CNX', 'target': [majorana_index_now, qubit_index], })
            sequence.append({'name': 'MIXTURE', 'target': [majorana_index_now, qubit_index], 'p': 0})

    ##  对于X型或Z型的探测
    elif type == 'x' or type == 'X' or type == 'z' or type == 'Z':

        ##  将qubit置于负号匹配
        sequence.append({'name': 'X', 'target': qubit_index})
        sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': 0})

        ##  生成前一半线路
        for j in range(len(occupy)):
            majorana_index_now = occupy[j]

            ##  最后一位与qubit作用CNX gate
            if j == len(occupy) - 1:
                sequence.append({'name': 'CNX', 'target': [majorana_index_now, qubit_index], })
                sequence.append({'name': 'MIXTURE', 'target': [majorana_index_now, qubit_index], 'p': 0})
                break

            majorana_index_down = occupy[j + 1]  # 后一个fermionic site

            ##  作用braid gate
            if flag:

                ##  根据稳定子类型选择braid形式
                if type == 'x' or type == 'X':
                    order_target = [majorana_index_down, majorana_index_now]
                elif type == 'z' or type == 'Z':
                    order_target = [majorana_index_now, majorana_index_down]
                else:
                    raise ValueError

                ##  添加braid gate
                sequence.append({"name": "BRAID", "target": order_target,})
                sequence.append({'name': 'FDEPOLARIZE2', 'target': order_target, 'p': 0})
                flag = False

            ##  作用CNN gate
            else:
                sequence.append({'name': 'CNN', 'target': [majorana_index_now, majorana_index_down], })
                sequence.append({'name': 'FDEPOLARIZE2', 'target': [majorana_index_now, majorana_index_down], 'p': 0})
                flag = True

        ##  生成syndrome extraction circuit的另一半
        flag = True
        for j in range(len(occupy) - 1):
            majorana_index_now = occupy[-1 - j]  # 当前的fermionic site
            majorana_index_up = occupy[-1 - j - 1]  # 上一个fermionic site

            ##  作用braid gate
            if flag:
                if type == 'x' or type == 'X':
                    order_target = [majorana_index_now, majorana_index_up]
                elif type == 'z' or type == 'Z':
                    order_target = [majorana_index_up, majorana_index_now]
                else:
                    raise ValueError
                sequence.append({'name': 'N', 'target': [majorana_index_now]})
                sequence.append({'name': 'BRAID', 'target': order_target})
                sequence.append({'name': 'N', 'target': [majorana_index_now]})
                sequence.append({'name': 'FDEPOLARIZE2', 'target': order_target, 'p': 0})
                flag = False

            ##  作用CNN gate
            else:
                sequence.append({'name': 'CNN', 'target': [majorana_index_now, majorana_index_up]})
                sequence.append({'name': 'FDEPOLARIZE2', 'target': [majorana_index_now, majorana_index_up], 'p': 0})
                flag = True

    ##  在qubit上测量结果并重置
    sequence.append({'name': 'MZ', 'target': qubit_index, 'p': 0})
    sequence.append({'name': 'R', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': 0})

    ##  SECTION：----返回线路序列----
    return sequence


# %%  CHAPTER：====生成Pauli CSS stabilizer测量线路====
def _syndrome_pauli_css_measurement_circuit(stabilizer: PauliOperator, qubit_index: int, type: str) -> list[dict]:
    """""
    input.stabilizer：一个PauliOperator，代表stabilizer
    input.qubit_index：一个整数，代表测量的qubit索引
    input.type：一个字符串，代表测量的类型，只能是'x'或'X'或'z'或'Z'
    input.p_noise：去极化噪声发生的几率
    input.p_measure：测量结果出错的几率
    """""

    ##  SECTION：----数据预处理----
    sequence = []  # 线路序列

    ##  判断稳定子类型
    if type == 'x' or type == 'X':
        occupy = stabilizer.occupy_x
    elif type == 'z' or type == 'Z':
        occupy = stabilizer.occupy_z
    else:
        raise ValueError

    ##  SECTION：----生成纠缠线路----
    if type == 'X' or type == 'x':
        sequence.append({'name': 'H', 'target': qubit_index})
        sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': 0})
    for j in range(len(occupy)):
        if type == 'Z' or type == 'z':
            sequence.append({'name': 'CX', 'target': [occupy[j], qubit_index]})
            sequence.append({'name': 'DEPOLARIZE2', 'target': [occupy[j], qubit_index], 'p': 0})
        elif type == 'X' or type == 'x':
            sequence.append({'name': 'CX', 'target': [qubit_index, occupy[j]]})
            sequence.append({'name': 'DEPOLARIZE2', 'target': [occupy[j], qubit_index], 'p': 0})
        else:
            raise ValueError
    if type == 'X' or type == 'x':
        sequence.append({'name': 'H', 'target': qubit_index})
        sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': 0})

    ##  在qubit上测量结果并重置
    sequence.append({'name': 'MZ', 'target': qubit_index, 'p': 0})
    sequence.append({'name': 'R', 'target': qubit_index})
    sequence.append({'name': 'DEPOLARIZE1', 'target': qubit_index, 'p': 0})

    ##  SECTION：----返回线路序列----
    return sequence



import copy
import networkx as nx
import numpy as np
from extendedstim.Code.BeyondCode.SubsystemCode import SubsystemCode
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.MajoranaCode import MajoranaCode
from extendedstim.Code.CodeFactory.ColorCode import MajoranaColorCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator


class FermionicLatticeSurgery:

    # %%  CHAPTER：====生成Majorana Lattice Surgery====
    def __init__(self,code_A:MajoranaCSSCode,index_A:int,code_B:MajoranaCSSCode|None=None,index_B:int|None=None)->None:
        """""
        构造用于测量joint logical observable的lattice surgery代码
        input.code_A：Majorana CSS Code
        input.index_A：code_A中的logical observable的索引
        input.code_B：Majorana CSS Code
        input.index_B：code_B中的logical observable的索引
        """""

        ##  PART：----初始化属性----
        ##  一开始的stabilizers
        self.observable:None|MajoranaOperator=None  # 待测的joint logical observable
        self.related_stabilizers_A_x=[]
        self.related_stabilizers_B_x=[]
        self.related_stabilizers_A_z=[]
        self.related_stabilizers_B_z=[]
        self.irrelated_stabilizers_A_z=[]
        self.irrelated_stabilizers_B_z=[]
        self.irrelated_stabilizers_A_x=[]
        self.irrelated_stabilizers_B_x=[]

        ##  添加的ancilla相关属性
        self.ancilla_number=0  # lattice surgery过程中生成的ancilla sites数目
        self.ancilla_stabilizers=[]  # 初始时刻ancilla stabilizers

        ##  需要生成的stabilizers
        self.modify_stabilizers_A=[]  # code_A中修改的stabilizer
        self.modify_stabilizers_B=[]  # code_B中修改的stabilizer
        self.gauge_stabilizers=[]  # lattice surgery生成的gauge stabilizers
        self.measurement_stabilizers=[]  # lattice surgery生成的measurement stabilizers

        ##  PART：----根据输入的情况不同使用两种不同的lattice surgery decoder_name----
        if index_B is None or code_B is None:
            self.code_A:MajoranaCSSCode=code_A
            self.index_A: int=index_A
            self.code_B: MajoranaColorCode=MajoranaColorCode(code_A.logical_operators_x[self.index_A].weight)
            self.index_B:int=0
            self._ZechuanLatticeSurgery()
        else:
            if code_A.logical_operators_x[index_A].weight>code_B.logical_operators_x[index_B].weight:
                self.code_A:MajoranaCSSCode=code_A.copy()
                self.code_B:MajoranaCSSCode=code_B.copy()
                self.index_A:int=index_A
                self.index_B:int=index_B
            else:
                self.code_A:MajoranaCSSCode=code_B.copy()
                self.code_B:MajoranaCSSCode=code_A.copy()
                self.index_A:int=index_B
                self.index_B:int=index_A
            self._MajoranaLatticeSurgery()
        self.physical_number=code_A.physical_number+self.code_B.physical_number+self.ancilla_number

        ##  PART：----从logical operators取出反对易于可观测量的算符生成subsystem code----
        self.code=MajoranaCode(self.irrelated_stabilizers_A_z+self.irrelated_stabilizers_B_z+self.irrelated_stabilizers_A_x+self.irrelated_stabilizers_B_x+
                               self.related_stabilizers_A_z+self.related_stabilizers_B_z+
                               self.modify_stabilizers_A+self.modify_stabilizers_B+self.gauge_stabilizers,self.physical_number)
        self.code_after_surgery=MajoranaCode(self.irrelated_stabilizers_A_x+self.irrelated_stabilizers_B_x+
                                             self.irrelated_stabilizers_A_z+self.irrelated_stabilizers_B_x+
                                             self.related_stabilizers_A_z+self.related_stabilizers_B_z+
                                             self.modify_stabilizers_A+self.modify_stabilizers_B+self.measurement_stabilizers+self.gauge_stabilizers,self.physical_number)

        ##  提取bare logical operators
        self.bare_logical_operators=[]
        for i in range(len(self.code_A.logical_operators_x)):
            self.bare_logical_operators.append(self.code_A.logical_operators_z[i])
            if i!=self.index_A:
                self.bare_logical_operators.append(self.code_A.logical_operators_x[i])
        for i in range(len(self.code_B.logical_operators_x)):
            self.bare_logical_operators.append(MajoranaOperator.HermitianOperatorFromOccupy([],self.code_B.logical_operators_z[i].occupy_z))
            if i!=self.index_B:
                self.bare_logical_operators.append(MajoranaOperator.HermitianOperatorFromOccupy(self.code_B.logical_operators_x[i].occupy_x,[]))

        self.subsystem_code=SubsystemCode(self.code, self.bare_logical_operators)

    # %%  CHAPTER：实现fermionic lattice surgery, decoder_name 1
    def _MajoranaLatticeSurgery(self):
        code_A:MajoranaCSSCode=self.code_A.copy()
        code_B:MajoranaCSSCode=self.code_B.copy()
        assert isinstance(code_A, MajoranaCSSCode)
        assert isinstance(code_B, MajoranaCSSCode)
        assert isinstance(self.index_A, int)
        assert isinstance(self.index_B, int)

        # %%  PART：----数据预处理----
        ##  匹配到joint code block的编号
        majorana_number_A=code_A.physical_number
        majorana_number_B=code_B.physical_number
        logical_operator_A=code_A.logical_operators_x[self.index_A]
        logical_operator_B=code_B.logical_operators_x[self.index_B]
        logical_operator_B.occupy_x=logical_operator_B.occupy_x+majorana_number_A
        self.observable=1j*logical_operator_A@logical_operator_B
        code_A.index_map(np.arange(majorana_number_A),majorana_number_A+majorana_number_B)
        code_B.index_map(np.arange(majorana_number_A, majorana_number_A+majorana_number_B),majorana_number_A+majorana_number_B)

        ##  获取需要修改的generators
        self.related_stabilizers_A_x=copy.deepcopy([temp for temp in code_A.generators if len(set(temp.occupy_x) & set(logical_operator_A.occupy_x)) > 0])
        self.related_stabilizers_B_x=copy.deepcopy([temp for temp in code_B.generators if len(set(temp.occupy_x) & set(logical_operator_B.occupy_x)) > 0])
        self.related_stabilizers_A_z=copy.deepcopy([temp for temp in code_A.generators_z if len(set(temp.occupy_z) & set(logical_operator_A.occupy_x)) > 0])
        self.related_stabilizers_B_z=copy.deepcopy([temp for temp in code_B.generators_z if len(set(temp.occupy_z) & set(logical_operator_B.occupy_x)) > 0])
        self.irrelated_stabilizers_A_z=copy.deepcopy([temp for temp in code_A.generators_z if len(set(temp.occupy_z) & set(logical_operator_A.occupy_x)) == 0])
        self.irrelated_stabilizers_B_z=copy.deepcopy([temp for temp in code_B.generators_z if len(set(temp.occupy_z) & set(logical_operator_B.occupy_x)) == 0])
        self.irrelated_stabilizers_A_x=copy.deepcopy([temp for temp in code_A.generators_x if len(set(temp.occupy_x) & set(logical_operator_A.occupy_x)) == 0])
        self.irrelated_stabilizers_B_x=copy.deepcopy([temp for temp in code_B.generators_x if len(set(temp.occupy_x) & set(logical_operator_B.occupy_x)) == 0])

        ##  PART：获取modified stabilizers
        ancilla_mode_number=0

        ##  为待修改的A stabilizers增加ancilla及其索引
        for i in range(len(self.related_stabilizers_A_x)):
            result_temp = [ancilla_mode_number + temp for temp in range(len(set(self.related_stabilizers_A_x[i].occupy_x) & set(logical_operator_A.occupy_x)))]  #  给出新增的ancilla modes编号
            self.modify_stabilizers_A.append([self.related_stabilizers_A_x[i].occupy_x.tolist(), result_temp])
            ancilla_mode_number+=len(result_temp)

        ##  为待修改的B stabilizers增加ancilla及其索引
        for i in range(len(self.related_stabilizers_B_x)):
            result_temp = [ancilla_mode_number + temp for temp in range(len(set(self.related_stabilizers_B_x[i].occupy_x) & set(logical_operator_B.occupy_x)))]  #  给出新增的ancilla modes编号
            self.modify_stabilizers_B.append([self.related_stabilizers_B_x[i].occupy_x.tolist(), result_temp])
            ancilla_mode_number+=len(result_temp)

        # %%  PART：加入测量稳定子，左边更长
        ##  先将两边对齐的部分连起来
        for i in range(len(logical_operator_B.occupy_x)):
            self.measurement_stabilizers.append([[logical_operator_A.occupy_x[i],logical_operator_B.occupy_x[i]],[]])
        for i in range((len(logical_operator_A.occupy_x)-len(logical_operator_B.occupy_x))//2):
            self.measurement_stabilizers.append([[logical_operator_A.occupy_x[2*i+len(logical_operator_B.occupy_x)],logical_operator_A.occupy_x[2*i+len(logical_operator_B.occupy_x)+1]],[]])

        ##  遍历A modified stabilizers，根据它与measurement stabilizer的交叉情况往measurement stabilizer上加ancilla modes
        for i in range(len(self.modify_stabilizers_A)):
            flag_temp=0  # 记录目前用到了哪个ancilla mode
            for j in range(len(self.measurement_stabilizers)):
                number=len(set(self.modify_stabilizers_A[i][0])&set(self.measurement_stabilizers[j][0]))
                self.measurement_stabilizers[j][1]=self.measurement_stabilizers[j][1]+[self.modify_stabilizers_A[i][1][flag_temp+k] for k in range(number)]
                flag_temp+=number

        ##  遍历B modified stabilizers，根据它与measurement stabilizer的交叉情况往measurement stabilizer上加ancilla modes
        for i in range(len(self.modify_stabilizers_B)):
            flag_temp=0  # 记录目前用到了哪个ancilla mode
            for j in range(len(self.measurement_stabilizers)):
                number=len(set(self.modify_stabilizers_B[i][0])&set(self.measurement_stabilizers[j][0]))
                self.measurement_stabilizers[j][1]=self.measurement_stabilizers[j][1]+[self.modify_stabilizers_B[i][1][flag_temp+k] for k in range(number)]
                flag_temp+=number

        ##  处理odd-weight measurement stabilizers
        single_mode_list=[]
        for i in range(len(self.measurement_stabilizers)):
            if len(self.measurement_stabilizers[i][1])%2!=0:
                self.measurement_stabilizers[i][1].append(ancilla_mode_number)
                single_mode_list.append(ancilla_mode_number)
                ancilla_mode_number+=1

        ##  PART：图论计算规范稳定子
        ##  初始化节点
        mode_vertices=['Mode'+str(temp) for temp in range(ancilla_mode_number)]
        A_vertices=[]
        B_vertices=[]
        measurement_vertices=[]
        edges=[]

        ##  生成A modified stabilizer nodes以及对应的edges
        for i in range(len(self.modify_stabilizers_A)):
            if len(self.modify_stabilizers_A[i][1])//2==1:
                A_vertices.append('A'+str(i))
                edges.append(('A'+str(i), 'Mode'+str(self.modify_stabilizers_A[i][1][0])))
                edges.append(('A'+str(i), 'Mode'+str(self.modify_stabilizers_A[i][1][1])))
            else:
                for j in range(len(self.modify_stabilizers_A[i][1])//2):
                    A_vertices.append('A'+str(i)+'_'+str(j))
                    edges.append(('A'+str(i), 'Mode'+str(self.modify_stabilizers_A[i][1][j*2])))
                    edges.append(('A'+str(i), 'Mode'+str(self.modify_stabilizers_A[i][1][j*2+1])))

        ##  生成B modified stabilizer nodes以及对应的edges
        for i in range(len(self.modify_stabilizers_B)):
            if len(self.modify_stabilizers_B[i][1])//2==1:
                B_vertices.append('B'+str(i))
                edges.append(('B'+str(i), 'Mode'+str(self.modify_stabilizers_B[i][1][0])))
                edges.append(('B'+str(i), 'Mode'+str(self.modify_stabilizers_B[i][1][1])))
            else:
                for j in range(len(self.modify_stabilizers_B[i][1])//2):
                    B_vertices.append('B'+str(i)+'_'+str(j))
                    edges.append(('B'+str(i), 'Mode'+str(self.modify_stabilizers_B[i][1][j*2])))
                    edges.append(('B'+str(i), 'Mode'+str(self.modify_stabilizers_B[i][1][j*2+1])))

        ##  生成measurement stabilizer nodes以及对应的edges
        for i in range(len(self.measurement_stabilizers)):
            measurement_vertices.append('Measure'+str(i))
            for j in range(len(self.measurement_stabilizers[i][1])):
                edges.append(('Measure'+str(i), 'Mode'+str(self.measurement_stabilizers[i][1][j])))

        ##  生成single ancilla mode的virtual edges
        virtual_vertices=[]
        for i in range(len(single_mode_list)//2):
            virtual_vertices.append('Virtual'+str(i))
            edges.append(('Virtual'+str(i), 'Mode'+str(single_mode_list[i])))
            edges.append(('Virtual'+str(i), 'Mode'+str(single_mode_list[i])))

        ##  生成图
        graph = nx.Graph()
        graph.add_nodes_from(mode_vertices)
        graph.add_nodes_from(A_vertices)
        graph.add_nodes_from(B_vertices)
        graph.add_nodes_from(measurement_vertices)
        graph.add_nodes_from(virtual_vertices)
        graph.add_edges_from(edges)
        cyclic_basis=nx.minimum_cycle_basis(graph)
        for i in range(len(cyclic_basis)):
            modes_temp=[]
            for j in range(len(cyclic_basis[i])):
                if 'Mode' in cyclic_basis[i][j]:
                    modes_temp.append(int(cyclic_basis[i][j][4:]))
            self.gauge_stabilizers.append([[],modes_temp])

        ##  PART：将stabilizers修改为正确的格式
        alls=[self.modify_stabilizers_A,self.modify_stabilizers_B,self.measurement_stabilizers,self.gauge_stabilizers]
        for i in range(len(alls)):
            for j in range(len(alls[i])):
                occupy_x_temp=alls[i][j][0]+[temp//2+majorana_number_A+majorana_number_B for temp in alls[i][j][1] if temp%2==0]
                occupy_z_temp=[temp//2+majorana_number_A+majorana_number_B for temp in alls[i][j][1] if temp%2==1]
                alls[i][j]=MajoranaOperator.HermitianOperatorFromOccupy(occupy_x_temp,occupy_z_temp)
        self.ancilla_number=ancilla_mode_number//2

    #%%  CHAPTER：Zechuan Lattice Surgery
    def _ZechuanLatticeSurgery(self):

        #%%  SECTION：----数据标准化----
        code_A: MajoranaCSSCode=self.code_A.copy()
        code_B: MajoranaColorCode=self.code_B.copy()
        majorana_number_A=code_A.physical_number
        majorana_number_B=code_B.physical_number
        logical_operator_A=code_A.logical_operators_x[self.index_A]
        logical_operator_B=code_B.logical_operators_x[0]
        logical_operator_B.occupy_x=logical_operator_B.occupy_x+majorana_number_A
        self.observable=1j*logical_operator_A@logical_operator_B
        code_A.index_map(np.arange(majorana_number_A),majorana_number_A+majorana_number_B)
        code_B.index_map(np.arange(majorana_number_A, majorana_number_A+majorana_number_B),majorana_number_A+majorana_number_B)

        self.related_stabilizers_A_x = copy.deepcopy([temp for temp in code_A.generators if len(set(temp.occupy_x) & set(logical_operator_A.occupy_x)) > 0])
        self.related_stabilizers_B_x = copy.deepcopy([temp for temp in code_B.generators if len(set(temp.occupy_x) & set(logical_operator_B.occupy_x)) > 0])
        self.irrelated_stabilizers_A_z = copy.deepcopy([temp for temp in code_A.generators_z if len(set(temp.occupy_z) & set(logical_operator_A.occupy_x)) == 0])
        self.irrelated_stabilizers_B_z = copy.deepcopy([temp for temp in code_B.generators_z if len(set(temp.occupy_z) & set(logical_operator_B.occupy_x)) == 0])
        self.irrelated_stabilizers_A_x = copy.deepcopy([temp for temp in code_A.generators_x if len(set(temp.occupy_x) & set(logical_operator_A.occupy_x)) == 0])
        self.irrelated_stabilizers_B_x = copy.deepcopy([temp for temp in code_B.generators_x if len(set(temp.occupy_x) & set(logical_operator_B.occupy_x)) == 0])

        ancilla_mode_number=0
        edge_site_bin_A:list[list[int]]=[[] for _ in range(code_B.d)]
        edge_site_bin_B:list[list[int]]=[[] for _ in range(code_B.d)]
        used_flags=[False for _ in range(len(code_B.edge_stabilizers))]

        ##  为待修改的stabilizers增加ancilla及其索引
        for i in range(len(self.related_stabilizers_A_x)):
            overlaps=list(set(self.related_stabilizers_A_x[i].occupy_x) & set(logical_operator_A.occupy_x))
            for j in range(len(overlaps)//2):
                index_0=int(np.where(logical_operator_A.occupy_x==overlaps[2*j])[0][0])
                index_1=int(np.where(logical_operator_A.occupy_x==overlaps[2*j+1])[0][0])
                edge_site_bin_A[index_0].append(ancilla_mode_number)
                edge_site_bin_A[index_1].append(ancilla_mode_number+1)
                self.modify_stabilizers_A.append([self.related_stabilizers_A_x[i].occupy_x.tolist(), [ancilla_mode_number, ancilla_mode_number + 1]])
                self.related_stabilizers_A_z.append(MajoranaOperator.HermitianOperatorFromOccupy([],self.related_stabilizers_A_x[i].occupy_x.tolist()))
                edge_site_bin_B[index_0].append(ancilla_mode_number+2)
                edge_site_bin_B[index_1].append(ancilla_mode_number+3)
                B_temp=self.code_B.get_between(index_0,index_1)
                self.related_stabilizers_B_z.append(MajoranaOperator.HermitianOperatorFromOccupy([], [temp+majorana_number_A for temp in B_temp.occupy_x]))
                self.modify_stabilizers_B.append([[temp+majorana_number_A for temp in B_temp.occupy_x], [ancilla_mode_number + 2, ancilla_mode_number + 3]])
                small=min(index_0,index_1)
                big=max(index_0,index_1)
                used_flags[small]=True
                used_flags[big-1]=True
                self.gauge_stabilizers.append([[],[ancilla_mode_number+temp for temp in range(4)]])
                ancilla_mode_number+=4

        ##  处理落单的edge stabilizers
        for i in range(len(used_flags)):
            if not used_flags[i]:
                edge_site_bin_B[i].append(ancilla_mode_number)
                edge_site_bin_B[i].append(ancilla_mode_number+1)
                edge_site_bin_B[i+1].append(ancilla_mode_number+2)
                edge_site_bin_B[i+1].append(ancilla_mode_number+3)
                self.modify_stabilizers_B.append([code_B.get_between(i, i+1).occupy_x.tolist(), [ancilla_mode_number, ancilla_mode_number+2]])
                self.gauge_stabilizers.append([[],[ancilla_mode_number+temp for temp in range(4)]])
                ancilla_mode_number+=4

        ##  添加measurement stabilizers
        for i in range(len(edge_site_bin_A)):
            self.measurement_stabilizers.append(([logical_operator_A.occupy_x[i], logical_operator_B.occupy_x[i]],edge_site_bin_A[i]+edge_site_bin_B[i]))

        ##  PART：将stabilizers修改为正确的格式
        alls=[self.modify_stabilizers_A,self.modify_stabilizers_B,self.measurement_stabilizers,self.gauge_stabilizers]
        for i in range(len(alls)):
            for j in range(len(alls[i])):
                occupy_x_temp=alls[i][j][0]+[temp//2+majorana_number_A+majorana_number_B for temp in alls[i][j][1] if temp%2==0]
                occupy_z_temp=[temp//2+majorana_number_A+majorana_number_B for temp in alls[i][j][1] if temp%2==1]
                alls[i][j]=MajoranaOperator.HermitianOperatorFromOccupy(occupy_x_temp,occupy_z_temp)
        self.ancilla_number=ancilla_mode_number//2


def get_between(color_code:MajoranaCSSCode,index_0:int,index_1:int):
    """""
    input.index_0：int，site的索引
    input.index_1：int，site的索引
    output：MajoranaOperator，index_0到index_1之间的stabilizer
    """""
    edge_stabilizers=color_code.generators_x[0:color_code.d]
    if index_0>index_1:
        aa=index_1
        bb=index_0
    else:
        aa=index_0
        bb=index_1
    assert 0<=index_0<color_code.d
    assert 0<=index_1<color_code.d
    if aa==bb-1:
        return edge_stabilizers[aa].copy()
    else:
        return edge_stabilizers[aa]@edge_stabilizers[bb-1]



"""""
模块作用：实现Majorana算符在占据表示下的代数运算、厄米性与向量化。
"""""
import galois
import numpy as np
from extendedstim.Physics.Operator import Operator
from extendedstim.tools.TypingTools import islist


class MajoranaOperator(Operator):
    __slots__ = ['occupy_x', 'occupy_z', 'coff']

    #%%  CHAPTER：====构造方法====
    def __init__(self, occupy_x, occupy_z, coff):
        """""
        input.occupy_x：γ支撑X索引
        input.occupy_z：γ'支撑Z索引
        input.coff：相位系数 (±1, ±i)
        """""
        super().__init__(occupy_x, occupy_z, coff)

    #%%  CHAPTER：====重载运算符====
    ##  SECTION：----矩阵乘法----
    def __matmul__(self, other:'MajoranaOperator')->'MajoranaOperator':

        ##  PART：----数据预处理----
        assert isinstance(other, MajoranaOperator)

        ##  PART：----计算新算符----
        ##  计算新的占据
        occupy_x = np.setxor1d(self.occupy_x, other.occupy_x, assume_unique=True)
        occupy_z = np.setxor1d(self.occupy_z, other.occupy_z, assume_unique=True)
        self_occupy = np.append(self.occupy_x * 2, self.occupy_z * 2 + 1)
        other_occupy = np.append(other.occupy_x * 2, other.occupy_z * 2 + 1)
        self_occupy = np.sort(self_occupy)
        other_occupy = np.sort(other_occupy)

        ##  计算新的系数
        exchange_times = np.sum([np.count_nonzero(self_occupy > temp) for temp in other_occupy])
        if exchange_times % 2 == 1:
            factor = -1
        else:
            factor = 1

        ##  PART：----返回新算符----
        return MajoranaOperator(occupy_x, occupy_z, self.coff * other.coff * factor)

    ##  SECTION：----右矩阵乘法----
    def __rmatmul__(self, other:'MajoranaOperator')->'MajoranaOperator':
        assert isinstance(other, MajoranaOperator)
        return other.__matmul__(self)

    ##  SECTION：----标量乘法----
    def __mul__(self, other:complex|float|int)->'MajoranaOperator':
        assert other == 1 or other == -1 or other == 1j or other == -1j
        return MajoranaOperator(self.occupy_x, self.occupy_z, self.coff * other)

    ##  SECTION：----右标量乘法----
    def __rmul__(self, other:complex|float|int)->'MajoranaOperator':
        return self.__mul__(other)

    ##  SECTION：----字符串表示----
    def __str__(self):
        return "MajoranaOperator(occupy_x={},occupy_z={},coff={})".format(self.occupy_x, self.occupy_z, self.coff)

    ##  SECTION：----相等判断----
    def __eq__(self, other:'MajoranaOperator')->bool:
        assert isinstance(other, MajoranaOperator)
        return np.array_equal(self.occupy_x, other.occupy_x) and np.array_equal(self.occupy_z, other.occupy_z) and self.coff == other.coff

    ##  SECTION：----取负----
    def __neg__(self)->'MajoranaOperator':
        return MajoranaOperator(self.occupy_x, self.occupy_z, -self.coff)

    # %%  CHAPTER：====属性方法====
    ##  SECTION：----算符是否是厄米算符----
    @property
    def is_hermitian(self)->bool:
        """""
        检查Majorana算符是否是厄米算符
        output：bool
        """""

        ##  PART：----计算厄米性质----
        if np.mod(self.weight * (self.weight - 1) // 2,2) == 0:
            if not (self.coff == 1 or self.coff == -1):
                return False
        else:
            if not (self.coff == 1j or self.coff == -1j):
                return False

        ##  PART：----返回结果----
        return True

    ##  SECTION：----求算符的对偶算符----
    @property
    def dual(self)->'MajoranaOperator':
        """""
        求Majorana算符的对偶算符
        output：MajoranaOperator
        """""
        return MajoranaOperator(self.occupy_z, self.occupy_x, self.coff)

    # %%  CHAPTER：====对象方法====
    ##  SECTION：----复制方法----
    def copy(self)->'MajoranaOperator':
        """""
        复制Majorana算符
        output：MajoranaOperator
        """""
        return MajoranaOperator(self.occupy_x.copy(), self.occupy_z.copy(), self.coff)

    # %%  CHAPTER：====静态方法====
    ##  SECTION：----定义一个厄米算符，从占据处表示----
    @staticmethod
    def HermitianOperatorFromOccupy(occupy_x:list[int]|np.ndarray,occupy_z:list[int]|np.ndarray)->'MajoranaOperator':
        """""
        将一个占据向量表示的Majorana算符转换为MajoranaOperator对象
        input.occupy_x：γ支撑X索引
        input.occupy_z：γ'支撑Z索引
        output：MajoranaOperator
        """""

        ##  PART：----数据预处理----
        assert islist(occupy_x)
        assert islist(occupy_z)

        ##  PART：----计算系数----
        weight = len(occupy_x) + len(occupy_z)
        if (weight * (weight - 1) // 2) % 2 == 0:
            coff = 1
        else:
            coff = 1j

        ##  PART：----返回结果----
        return MajoranaOperator(occupy_x, occupy_z, coff)

    ##  SECTION：----定义一个厄米算符，从向量表示----
    @staticmethod
    def HermitianOperatorFromVector(vector)->'MajoranaOperator':
        """""
        将一个binary vector转换为Majorana算符
        input.vector：Majorana算符的向量表示
        output：MajoranaOperator
        """""
        ##  PART：----数据预处理----
        assert isinstance(vector, (np.ndarray, galois.GF(2)))

        ##  PART：----计算占据----
        occupy_x = np.where(vector[0::2] == 1)[0]
        occupy_z = np.where(vector[1::2] == 1)[0]

        ##  PART：----返回结果----
        return MajoranaOperator.HermitianOperatorFromOccupy(occupy_x, occupy_z)

    ##  SECTION：----检查两个厄米算符是否对易----
    @staticmethod
    def commute(A:'MajoranaOperator', B:'MajoranaOperator')->bool:
        """""
        input.A：MajoranaOperator
        input.B：MajoranaOperator
        output：bool，对易返回True
        """""
        assert isinstance(A, MajoranaOperator) and isinstance(B, MajoranaOperator)
        overlap_x = len(np.intersect1d(A.occupy_x, B.occupy_x))
        overlap_z = len(np.intersect1d(A.occupy_z, B.occupy_z))
        weight = (len(A.occupy_x) + len(A.occupy_z)) * (len(B.occupy_x) + len(B.occupy_z))
        judge = overlap_x + overlap_z + weight
        return np.mod(judge, 2) == 0


import galois
from tqdm.contrib.itertools import product
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.PauliCSSCode import PauliCSSCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator
from extendedstim.tools.TypingTools import islist, isinteger
import numpy as np


class PolynomialMatrix:
    def __init__(self, dimension,lengths,cell_size:int,polynomials):
        """""
        input：
            dimension: 维度
            lengths: 每个维度的长度
            cell_size: 每个unit cell中的qubit数量
            polynomials: 多项式描述，结构为 A[q][i][j]
                q: 代表unit cell中的第q个qubit (0 <= q < cell_size)
                i: 代表该qubit在stabilizer中出现的第i个位置 (即多项式中的第i项)
                j: 代表偏移向量的维度索引 [x, y, z...]
        """""
        assert len(polynomials)==cell_size
        assert islist(lengths) and all([isinteger(length) for length in lengths])
        
        # 规范化多项式中的偏移量 (mod lengths)
        for q in range(len(polynomials)): # q: qubit index within cell
            for i in range(len(polynomials[q])): # i: term index in polynomial
                polynomials[q][i]=np.mod(polynomials[q][i],lengths)
                
        self.polynomial = polynomials
        self.dimension = dimension
        self.cell_size = cell_size
        self.lengths = lengths
        self.physical_number=cell_size*np.prod(lengths)
        
        stabilizers=[]
        # 遍历整个晶格中的每一个位置 vector (作为stabilizer的中心位置)
        for vector in product(*[range(length) for length in lengths]):
            stabilizer_temp = np.zeros((cell_size,)+tuple(lengths))
            # index 对应 q (qubit type in cell)
            # polynomial 对应 A[q] (list of offsets for this qubit type)
            for q, offsets in enumerate(self.polynomial):
                for pos in offsets: # pos is one offset vector [x, y, z]
                    # Stabilizer中心在 vector，连接到的qubit位置是 vector + pos
                    vector_temp=np.array(vector,dtype=int)+np.array(pos,dtype=int)
                    # 标记对应的qubit (第q种qubit，位置为 vector_temp % lengths)
                    stabilizer_temp[q][*np.mod(vector_temp,lengths)]=1
            stabilizers.append(stabilizer_temp)
        stabilizers=np.array(stabilizers,dtype=int)
        stabilizers=galois.GF2(stabilizers.reshape((self.physical_number//cell_size,self.physical_number)))
        self.H=stabilizers

    #%%  SECTION：----重载函数----
    def __str__(self)->str:
        """""
        input：None
        output：字符串表示
        influence：返回多项式矩阵的字符串描述
        """""
        if self.dimension <= 3:
            vars = ['x', 'y', 'z']
        else:
            vars = [f'x_{i}' for i in range(self.dimension)]
            
        poly_strs = []
        for q in range(len(self.polynomial)):
            terms = []
            offsets = self.polynomial[q]
            if not offsets:
                terms.append("0")
            else:
                for pos in offsets:
                    term_parts = []
                    for dim_idx, p in enumerate(pos):
                        if p != 0:
                            if p == 1:
                                term_parts.append(f"{vars[dim_idx]}")
                            else:
                                term_parts.append(f"{vars[dim_idx]}^{p}")
                    if not term_parts:
                        terms.append("1")
                    else:
                        terms.append("".join(term_parts))
            poly_strs.append(" + ".join(terms) if terms else "0")
            
        return f"({', '.join(poly_strs)})"


class PolynomialCode:
    def __init__(self, h_x: PolynomialMatrix, h_z: PolynomialMatrix):
        assert h_x.dimension == h_z.dimension
        assert h_x.lengths == h_z.lengths
        assert h_x.cell_size == h_z.cell_size
        self.h_x = h_x
        self.h_z = h_z
        self.dimension = h_x.dimension
        self.lengths = h_x.lengths
        self.cell_size = h_x.cell_size
        self.H_X=h_x.H
        self.H_Z=h_z.H

    def _get_visual_pos(self, target, center, length):
        """Helper to get the nearest visual position accounting for PBC."""
        diff = target - center
        # Map diff to [-length/2, length/2]
        diff = (diff + length/2) % length - length/2
        return center + diff

    def _get_qubit_offset(self, q, cell_size):
        """
        计算cell中第q个qubit的视觉偏移量。
        排列规则：从左往右，从下往上 (Grid layout)
        """
        if cell_size == 1:
            return 0, 0
        
        # Grid layout dimensions
        nx = int(np.ceil(np.sqrt(cell_size)))
        ny = int(np.ceil(cell_size / nx))
        
        # Grid coordinates (col, row)
        col = q % nx
        row = q // nx
        
        # Map to visual offsets centered at (0,0)
        # spacing = 0.2
        spacing = 0.2
        
        # Center the grid
        # x range: [-(nx-1)/2, (nx-1)/2] * spacing
        dx = (col - (nx - 1) / 2) * spacing
        dy = (row - (ny - 1) / 2) * spacing
        
        return dx, dy

    def _get_qubit_style(self, q, cell_size):
        """
        获取qubit的颜色样式。
        使用低饱和度高级配色。
        """
        # 配色表 (fill, edge)
        # 0: Green (#86B257, #74A742)
        # 1: Blue (#87CEEB, #4682B4) - SkyBlue / SteelBlue
        # 2: Purple (#D8BFD8, #DA70D6) - Thistle / Orchid
        # 3: Orange (#FFDAB9, #FFA07A) - PeachPuff / LightSalmon
        
        colors = [
            ('#86B257', '#74A742'), # Green
            ('#87CEEB', '#4682B4'), # Blue
            ('#D8BFD8', '#DA70D6'), # Purple
            ('#FFDAB9', '#FFA07A'), # Orange
            ('#F0E68C', '#BDB76B'), # Khaki
            ('#E0FFFF', '#00CED1')  # LightCyan / DarkTurquoise
        ]
        
        fill, edge = colors[q % len(colors)]
        return fill, edge

    #%%  SECTION：----重载函数----
    def __str__(self)->str:
        """""
        input：None
        output：字符串表示
        influence：返回PolynomialCode的详细描述
        """""
        info = []
        info.append(f"Polynomial Code (Dimension: {self.dimension})")
        info.append(f"Lengths: {self.lengths}")
        info.append(f"Cell Size: {self.cell_size}")
        info.append(f"H_X Polynomials: {self.h_x}")
        info.append(f"H_Z Polynomials: {self.h_z}")
        return "\n".join(info)

    def _get_visual_pos_3d(self, tx, ty, tz, cx, cy, cz, Lx, Ly, Lz):
        dx = (tx - cx + Lx/2) % Lx - Lx/2
        dy = (ty - cy + Ly/2) % Ly - Ly/2
        dz = (tz - cz + Lz/2) % Lz - Lz/2
        return cx+dx, cy+dy, cz+dz

    #%%  SECTION：----绘图函数----
    def draw(self)->None:
        """""
        input：None
        output：None
        influence：绘制量子码的结构图，包括qubit的位置、正方晶格、以及X和Z stabilizer的结构。
        坐标系说明：
        - polynomial中的[x, y, z]分别对应[横坐标, 纵坐标, 层坐标]
        - 绘图时，x轴为水平向右，y轴为垂直向上(2D)或深度(3D)，z轴为垂直堆叠(3D)
        - Stabilizer绘制在坐标原点(0,0,...)所在的晶格中心
        """""
        try:
            import matplotlib.pyplot as plt
            import matplotlib.patches as patches
            from mpl_toolkits.mplot3d import Axes3D
        except ImportError:
            print("Matplotlib not found.")
            return
        
        # Stabilizer Colors
        color_x = '#DD705D'
        color_z = '#5874B5'

        ##  PART：----一维情况----
        if self.dimension == 1:
            fig, ax = plt.subplots(figsize=(max(6, self.lengths[0]), 3))
            L = self.lengths[0]
            
            ##  绘制正方晶格（一维为线）
            ax.plot([-0.5, L-0.5], [0, 0], color='lightgray', linestyle='-', linewidth=2, zorder=0)
            ax.plot([L-1, L-0.5], [0, 0], color='lightgray', linestyle='--', zorder=0)
            ax.plot([-0.5, 0], [0, 0], color='lightgray', linestyle='--', zorder=0)

            ##  绘制Cell标识和Qubit
            for x in range(L):
                # 绘制Cell椭圆
                ellipse = patches.Ellipse((x, 0), width=0.6, height=0.3, 
                                        edgecolor='gray', facecolor='none', 
                                        linestyle='--', alpha=0.3, zorder=1)
                ax.add_patch(ellipse)
                
                for i in range(self.cell_size):
                    dx, dy = self._get_qubit_offset(i, self.cell_size)
                    fill_c, edge_c = self._get_qubit_style(i, self.cell_size)
                    # 缩小Qubit圆圈 (radius 0.1 -> 0.05?)
                    circle = plt.Circle((x + dx, dy), 0.06, facecolor=fill_c, edgecolor=edge_c, linewidth=1.5, zorder=10)
                    ax.add_patch(circle)

            ##  绘制Stabilizers
            cx = 0
            
            # X Stabilizer (Red)
            check_vis_x = cx + 0.5
            check_vis_y = 0.2
            ax.scatter([check_vis_x], [check_vis_y], marker='s', color=color_x, s=60, label='X Check', zorder=15)

            for q, offsets in enumerate(self.h_x.polynomial):
                for pos in offsets:
                    tx = (cx + pos[0]) % L
                    qubit_offset_x, qubit_offset_y = self._get_qubit_offset(q, self.cell_size)
                    vis_tx = self._get_visual_pos(tx, cx, L)
                    ax.plot([check_vis_x, vis_tx + qubit_offset_x], [check_vis_y, qubit_offset_y], color=color_x, linestyle='-', alpha=0.6, zorder=5)
            
            # Z Stabilizer (Blue)
            check_vis_x = cx + 0.5
            check_vis_y = -0.2
            ax.scatter([check_vis_x], [check_vis_y], marker='s', color=color_z, s=60, label='Z Check', zorder=15)

            for q, offsets in enumerate(self.h_z.polynomial):
                for pos in offsets:
                    tx = (cx + pos[0]) % L
                    qubit_offset_x, qubit_offset_y = self._get_qubit_offset(q, self.cell_size)
                    vis_tx = self._get_visual_pos(tx, cx, L)
                    ax.plot([check_vis_x, vis_tx + qubit_offset_x], [check_vis_y, qubit_offset_y], color=color_z, linestyle='-', alpha=0.6, zorder=5)

            ax.set_xlim(-1, L)
            ax.set_ylim(-1, 1)
            ax.set_aspect('equal')
            ax.axis('off')
            plt.title(f"1D Polynomial Code (PBC), L={L}")
            plt.legend(loc='upper right')
            plt.show()

        ##  PART：----二维情况----
        elif self.dimension == 2:
            fig, ax = plt.subplots(figsize=(6, 6))
            Lx, Ly = self.lengths
            
            ##  绘制正方晶格
            for y in range(Ly):
                ax.plot([0, Lx-1], [y, y], color='lightgray', linestyle='-', zorder=0)
            for x in range(Lx):
                ax.plot([x, x], [0, Ly-1], color='lightgray', linestyle='-', zorder=0)

            ##  绘制Cell标识和Qubit
            for x in range(Lx):
                for y in range(Ly):
                    # 绘制Cell椭圆 (圆形)
                    ellipse = patches.Ellipse((x, y), width=0.6, height=0.6, 
                                            edgecolor='gray', facecolor='none', 
                                            linestyle='--', alpha=0.3, zorder=1)
                    ax.add_patch(ellipse)
                    
                    for i in range(self.cell_size):
                        dx, dy = self._get_qubit_offset(i, self.cell_size)
                        fill_c, edge_c = self._get_qubit_style(i, self.cell_size)
                        circle = plt.Circle((x+dx, y+dy), 0.06, facecolor=fill_c, edgecolor=edge_c, linewidth=1.5, zorder=10)
                        ax.add_patch(circle)

            ##  绘制Stabilizers
            cx, cy = 0, 0
            
            # X Lines (Red)
            # One check marker at center for X, slightly offset
            check_vis_x, check_vis_y = cx + 0.5 - 0.1, cy + 0.5 - 0.1
            ax.scatter([check_vis_x], [check_vis_y], marker='s', color=color_x, s=60, label='X Check', zorder=15)

            for q, offsets in enumerate(self.h_x.polynomial):
                for pos in offsets:
                    # pos[0] -> x (horizontal), pos[1] -> y (vertical)
                    tx, ty = (cx + pos[0]) % Lx, (cy + pos[1]) % Ly
                    
                    # Qubit offset
                    dx, dy = self._get_qubit_offset(q, self.cell_size)
                    
                    vis_tx = self._get_visual_pos(tx, cx, Lx)
                    vis_ty = self._get_visual_pos(ty, cy, Ly)
                    
                    ax.plot([check_vis_x, vis_tx+dx], [check_vis_y, vis_ty+dy], color=color_x, linestyle='-', alpha=0.6, linewidth=1.5, zorder=5)

            # Z Lines (Blue)
            # One check marker at center for Z, slightly offset
            check_vis_x, check_vis_y = cx + 0.5 + 0.1, cy + 0.5 + 0.1
            ax.scatter([check_vis_x], [check_vis_y], marker='s', color=color_z, s=60, label='Z Check', zorder=15)

            for q, offsets in enumerate(self.h_z.polynomial):
                for pos in offsets:
                    tx, ty = (cx + pos[0]) % Lx, (cy + pos[1]) % Ly
                    
                    dx, dy = self._get_qubit_offset(q, self.cell_size)
                    
                    vis_tx = self._get_visual_pos(tx, cx, Lx)
                    vis_ty = self._get_visual_pos(ty, cy, Ly)
                    
                    ax.plot([check_vis_x, vis_tx+dx], [check_vis_y, vis_ty+dy], color=color_z, linestyle='-', alpha=0.6, linewidth=1.5, zorder=5)


            ##  体现PBC
            for y in range(Ly):
                ax.plot([Lx-1, Lx-0.5], [y, y], color='lightgray', linestyle='--')
                ax.plot([-0.5, 0], [y, y], color='lightgray', linestyle='--')
            for x in range(Lx):
                ax.plot([x, x], [Ly-1, Ly-0.5], color='lightgray', linestyle='--')
                ax.plot([x, x], [-0.5, 0], color='lightgray', linestyle='--')

            ax.set_aspect('equal')
            ax.axis('off')
            plt.title(f"2D Polynomial Code (PBC), Size={Lx}x{Ly}")
            plt.legend(loc='upper right')
            plt.show()

        ##  PART：----三维情况----
        elif self.dimension == 3:
            Lx, Ly, Lz = self.lengths
            # Create subplots: one for each z-layer
            fig, axes = plt.subplots(1, Lz, figsize=(4*Lz, 5), constrained_layout=True)
            if Lz == 1:
                axes = [axes]
            
            # Stabilizer center at (0,0,0)
            cx, cy, cz = 0, 0, 0
            
            # Visual positions for X and Z checks (slightly offset)
            check_vis_x_X, check_vis_y_X = cx + 0.5 - 0.1, cy + 0.5 - 0.1
            check_vis_x_Z, check_vis_y_Z = cx + 0.5 + 0.1, cy + 0.5 + 0.1

            for z_layer in range(Lz):
                ax = axes[z_layer]
                ax.set_title(f"Layer z={z_layer}")
                
                ##  绘制正方晶格 (2D Grid for this layer)
                for y in range(Ly):
                    ax.plot([0, Lx-1], [y, y], color='lightgray', linestyle='-', zorder=0)
                for x in range(Lx):
                    ax.plot([x, x], [0, Ly-1], color='lightgray', linestyle='-', zorder=0)

                ##  绘制Cell标识和Qubit
                for x in range(Lx):
                    for y in range(Ly):
                        # Cell ellipse
                        ellipse = patches.Ellipse((x, y), width=0.6, height=0.6, 
                                                edgecolor='gray', facecolor='none', 
                                                linestyle='--', alpha=0.3, zorder=1)
                        ax.add_patch(ellipse)
                        
                        for i in range(self.cell_size):
                            dx, dy = self._get_qubit_offset(i, self.cell_size)
                            fill_c, edge_c = self._get_qubit_style(i, self.cell_size)
                            circle = plt.Circle((x+dx, y+dy), 0.06, facecolor=fill_c, edgecolor=edge_c, linewidth=1.5, zorder=10)
                            ax.add_patch(circle)

                ##  绘制Stabilizers
                
                # Check Markers
                # Only solid if on this layer, else ghost
                if z_layer == cz:
                    ax.scatter([check_vis_x_X], [check_vis_y_X], marker='s', color=color_x, s=60, label='X Check', zorder=15)
                    ax.scatter([check_vis_x_Z], [check_vis_y_Z], marker='s', color=color_z, s=60, label='Z Check', zorder=15)
                else:
                    # Ghost markers
                    ax.scatter([check_vis_x_X], [check_vis_y_X], marker='s', color=color_x, s=60, alpha=0.15, zorder=5)
                    ax.scatter([check_vis_x_Z], [check_vis_y_Z], marker='s', color=color_z, s=60, alpha=0.15, zorder=5)

                # X Connections
                for q, offsets in enumerate(self.h_x.polynomial):
                    for pos in offsets:
                        tx, ty, tz = (cx + pos[0]) % Lx, (cy + pos[1]) % Ly, (cz + pos[2]) % Lz
                        
                        if tz == z_layer:
                            dx, dy = self._get_qubit_offset(q, self.cell_size)
                            vis_tx = self._get_visual_pos(tx, cx, Lx)
                            vis_ty = self._get_visual_pos(ty, cy, Ly)
                            ax.plot([check_vis_x_X, vis_tx+dx], [check_vis_y_X, vis_ty+dy], color=color_x, linestyle='-', alpha=0.6, linewidth=1.5, zorder=10)

                # Z Connections
                for q, offsets in enumerate(self.h_z.polynomial):
                    for pos in offsets:
                        tx, ty, tz = (cx + pos[0]) % Lx, (cy + pos[1]) % Ly, (cz + pos[2]) % Lz
                        
                        if tz == z_layer:
                            dx, dy = self._get_qubit_offset(q, self.cell_size)
                            vis_tx = self._get_visual_pos(tx, cx, Lx)
                            vis_ty = self._get_visual_pos(ty, cy, Ly)
                            ax.plot([check_vis_x_Z, vis_tx+dx], [check_vis_y_Z, vis_ty+dy], color=color_z, linestyle='-', alpha=0.6, linewidth=1.5, zorder=10)

                ##  体现PBC
                for y in range(Ly):
                    ax.plot([Lx-1, Lx-0.5], [y, y], color='lightgray', linestyle='--')
                    ax.plot([-0.5, 0], [y, y], color='lightgray', linestyle='--')
                for x in range(Lx):
                    ax.plot([x, x], [Ly-1, Ly-0.5], color='lightgray', linestyle='--')
                    ax.plot([x, x], [-0.5, 0], color='lightgray', linestyle='--')

                ax.set_aspect('equal')
                ax.axis('off')
                
            # Global Legend (taking handles from first plot if available, or creating custom)
            handles = [
                plt.Line2D([0], [0], marker='s', color='w', markerfacecolor=color_x, markersize=10, label='X Check'),
                plt.Line2D([0], [0], marker='s', color='w', markerfacecolor=color_z, markersize=10, label='Z Check')
            ]
            fig.legend(handles=handles, loc='upper right')
            plt.suptitle(f"3D Polynomial Code Layers (L={Lx}x{Ly}x{Lz})")
            plt.show()


# %%  CHAPTER：====Majorana finite projective geometry code====
class MajoranaPolynomialCode(MajoranaCSSCode,PolynomialCode):

    ##  SECTION：----构造方法----
    def __init__(self,h_x:PolynomialMatrix,h_z:PolynomialMatrix):
        PolynomialCode.__init__(self,h_x,h_z)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromOccupy(np.where(self.H_X[i]!=0)[0],[]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromOccupy([],np.where(self.H_Z[i]!=0)[0]))
        MajoranaCSSCode.__init__(self,generators_x,generators_z,self.H.shape[1])


# %%  CHAPTER：====Pauli finite projective geometry code====
class PauliPolynomialCode(PauliCSSCode,PolynomialCode):

    ##  SECTION：----构造方法----
    def __init__(self,h_x:PolynomialMatrix,h_z:PolynomialMatrix):
        PolynomialCode.__init__(self,h_x,h_z)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(PauliOperator.HermitianOperatorFromOccupy(np.where(self.H_X[i]!=0)[0],[]))
            generators_z.append(PauliOperator.HermitianOperatorFromOccupy([],np.where(self.H_Z[i]!=0)[0]))
        PauliCSSCode.__init__(self,generators_x,generators_z,self.H.shape[1])


import numpy as np
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.PauliCSSCode import PauliCSSCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator
import galois


#%%  CHAPTER：====构造Twisted Double Bivariant Bicycle Code====
class TwistedDoubleBivariantBicycleCode:
    __slots__=['gamma','l','m','a_x_list','a_y_list','b_x_list','b_y_list','H']

    def __init__(self,l:int,m:int,gamma:int,a_x_list:list[int],a_y_list:list[int],b_x_list:list[int],b_y_list:list[int]):
        """""
        input.l：循环矩阵的大小 (偶数)
        input.m：循环矩阵的大小 (偶数)
        input.a_x_list：A中x多项式的幂列表
        input.a_y_list：A中y多项式的幂列表
        input.b_x_list：B中x多项式的幂列表
        input.b_y_list：B中y多项式的幂列表
        """""

        self.l=l
        self.m=m
        self.gamma=np.mod(gamma,m)
        self.a_x_list=a_x_list
        self.a_y_list=a_y_list
        self.b_x_list=b_x_list
        self.b_y_list=b_y_list
        GF=galois.GF(2)
        x=GF(np.kron(shift(l, 1), np.eye(m, dtype=int)))
        x[m*(l-1):,0:m]=shift(m,self.gamma)
        y=GF(np.kron(np.eye(l, dtype=int), shift(m, 1)))
        A=None
        B=None
        for i in range(len(a_x_list)):
            if A is None:
                A=np.linalg.matrix_power(x, a_x_list[i])@np.linalg.matrix_power(y, a_y_list[i])
            else:
                A=A+np.linalg.matrix_power(x, a_x_list[i])@np.linalg.matrix_power(y, a_y_list[i])
        for i in range(len(b_x_list)):
            if B is None:
                B=np.linalg.matrix_power(x, b_x_list[i])@np.linalg.matrix_power(y, b_y_list[i])
            else:
                B=B+np.linalg.matrix_power(x, b_x_list[i])@np.linalg.matrix_power(y, b_y_list[i])
        A_T=A.T
        B_T=B.T
        H_top=np.hstack((A, B, A_T, B_T))
        H_bottom=np.hstack((B_T, A_T, B, A))
        H=GF(np.vstack((H_top, H_bottom)))
        self.H=H

    ##  SECTION：----字符串表示----
    def __str__(self):
        A_str=''
        B_str=''
        for i in range(len(self.a_x_list)):
            if i==0:
                A_str=A_str+f"x^{self.a_x_list[i]}y^{self.a_y_list[i]}"
            else:
                A_str=A_str+f"+x^{self.a_x_list[i]}y^{self.a_y_list[i]}"
        for i in range(len(self.b_x_list)):
            if i==0:
                B_str=B_str+f"x^{self.b_x_list[i]}y^{self.b_y_list[i]}"
            else:
                B_str=B_str+f"+x^{self.b_x_list[i]}y^{self.b_y_list[i]}"
        return f"|{self.l*self.m*4}|{self.l}|{self.m}|{self.gamma}|"+A_str+"|"+B_str+"|"


# %%  CHAPTER：====Majorana twisted double bivariant bicycle code====
class MajoranaTwistedDoubleBivariantBicycleCode(MajoranaCSSCode,TwistedDoubleBivariantBicycleCode):

    ##  SECTION：----构造方法----
    def __init__(self,l:int,m:int,gamma:int,a_x_list:list[int],a_y_list:list[int],b_x_list:list[int],b_y_list:list[int]):
        TwistedDoubleBivariantBicycleCode.__init__(self,l,m,gamma,a_x_list,a_y_list,b_x_list,b_y_list)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromOccupy([],np.where(self.H[i]!=0)[0]))
        MajoranaCSSCode.__init__(self,generators_x,generators_z,self.l*self.m*4)


# %%  CHAPTER：====Pauli twisted double bivariant bicycle code====
class PauliTwistedDoubleBivariantBicycleCode(PauliCSSCode,TwistedDoubleBivariantBicycleCode):

    ##  SECTION：----构造方法----
    def __init__(self,l:int,m:int,gamma:int,a_x_list:list[int],a_y_list:list[int],b_x_list:list[int],b_y_list:list[int]):
        TwistedDoubleBivariantBicycleCode.__init__(self,l,m,gamma,a_x_list,a_y_list,b_x_list,b_y_list)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(PauliOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(PauliOperator.HermitianOperatorFromOccupy([],np.where(self.H[i]!=0)[0]))
        PauliCSSCode.__init__(self,generators_x,generators_z,self.l*self.m*4)


# %%  CHAPTER：====用到的静态函数====
##  SECTION：----生成循环移位矩阵----
def shift(number:int, shift_number:int):
    """""
    input.number：矩阵维度
    input.shift：循环移位步长
    output：GF(2)矩阵，置换矩阵
    """""

    S = np.zeros((number, number), dtype=int)
    for i in range(number):
        S[i, (i + shift_number) % number] = 1
    return galois.GF(2)(S)


"""""
模块作用：定义Pauli帧（Pauli Frame）模型，用于在Clifford电路模拟中跟踪误差传播，是`Platform`的简化/快速版本。
"""""
import galois
import numpy as np
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator
from extendedstim.tools.TypingTools import isinteger


class Frame:
    GF=galois.GF(2)

    # %%  CHAPTER：====构造方法====
    def __init__(self):
        """""
        influence：初始化空帧，比特数和费米子位数为0。
        """""
        self.pauli_number=0
        self.majorana_number=0
        self.frame=None
        self.pauli_frame=None
        self.majorana_frame=None

    # %%  CHAPTER：====对象方法====
    ##  SECTION：----强制初始化----
    def trap(self, majorana_number,pauli_number):
        """""
        input.majorana_state：Majorana稳定子列表
        input.pauli_state：Pauli稳定子列表
        influence：根据给定的稳定子随机初始化帧（模拟投影到随机的+1/-1本征态）。
        """""
        self.pauli_number=pauli_number
        self.majorana_number=majorana_number
        self.frame=self.GF.Zeros(2*pauli_number+2*majorana_number)
        for i in range(majorana_number):
            if np.random.rand()>0.5:
                temp=self.GF(np.zeros(2*majorana_number,dtype=int))
                temp[i*2]=1
                temp[i*2+1]=1
                self.frame[0:self.majorana_number*2]+=temp
        for i in range(pauli_number):
            if np.random.rand()>0.5:
                temp=self.GF(np.zeros(2*pauli_number,dtype=int))
                temp[i*2+1]=1
                self.frame[self.majorana_number*2:]+=temp

    ##  SECTION：----测量算符op，返回测量结果，随机坍缩----
    def measure(self, op,reference_value):
        """""
        input.op：待测量的厄米算符
        input.reference_value：无噪声参考测量值 (+1或-1)
        output：int，实际测量结果 (+1或-1)
        influence：根据算符与当前帧的对易/反对易关系翻转参考值，并以50%概率更新帧。
        """""

        ##  PART：----数据预处理----
        assert op.is_hermitian, "输入的厄米算符必须是厄米算符"
        assert reference_value==1 or reference_value==-1, "参考值必须是+1或-1"

        ##  PART：----测量算符op，返回测量结果，随机坍缩----
        if isinstance(op, MajoranaOperator):
            v0=op.get_vector(self.majorana_number)
            v1=self.frame[0:self.majorana_number*2]
            overlap_number=np.dot(v0,v1)
            weight_0=np.dot(v0,v0)
            weight_1=np.dot(v1,v1)
            if overlap_number+weight_0*weight_1==0:
                result=reference_value
            else:
                result=-reference_value
            if np.random.rand()>0.5:
                self.frame[0:self.majorana_number*2]+=v0
            return result
        elif isinstance(op, PauliOperator):
            v0=op.get_vector(self.pauli_number)
            v0_x=v0[0::2]
            v0_z=v0[1::2]
            v1=self.frame[self.majorana_number*2:]
            v1_x=v1[0::2]
            v1_z=v1[1::2]
            overlap_number=np.dot(v0_x,v1_z)+np.dot(v0_z,v1_x)
            if overlap_number==0:
                result=reference_value
            else:
                result=-reference_value
            if np.random.rand()>0.5:
                self.frame[self.majorana_number*2:]+=v0
            return result
        else:
            raise ValueError

    ##  SECTION：----X门，作用于qubit_index----
    def x(self, qubit_index: int):
        """""
        input.qubit_index：目标量子位
        influence：在Pauli帧模型中，理想的单比特门不改变误差帧，因此为空操作。
        """""
        pass

    ##  SECTION：----Y门，作用于qubit_index----
    def y(self, qubit_index: int):
        """""
        input.qubit_index：目标量子位
        influence：理想门，空操作。
        """""
        pass

    ##  SECTION：----Z门，作用于qubit_index----
    def z(self, qubit_index: int):
        """""
        input.qubit_index：目标量子位
        influence：理想门，空操作。
        """""
        pass

    ##  SECTION：----Hadamard gate，作用于qubit_index----
    def h(self, qubit_index: int):
        """""
        input.qubit_index：目标量子位
        influence：交换Pauli帧中对应量子位的X和Z分量。
        """""
        ##  PART：----数据预处理----
        assert isinteger(qubit_index) and 0<=qubit_index<self.pauli_number
        qubit_index_x=self.majorana_number*2+qubit_index*2
        qubit_index_z=self.majorana_number*2+qubit_index*2+1

        ##  PART：----交换----
        cache=self.frame[qubit_index_x].copy()
        self.frame[qubit_index_x]=self.frame[qubit_index_z]
        self.frame[qubit_index_z]=cache

    ##  SECTION：----S门，作用于pauli_index----
    def s(self, pauli_index: int):
        """""
        input.pauli_index：目标量子位
        influence：在Pauli帧中，Z分量加上X分量 (Z -> ZX)。
        """""

        ##  PART：----数据预处理----
        assert isinteger(pauli_index) and 0<=pauli_index<self.pauli_number
        qubit_index_x=self.majorana_number*2+pauli_index*2
        qubit_index_z=self.majorana_number*2+pauli_index*2+1

        ##  PART：----S门作用----
        self.frame[qubit_index_z]+=self.frame[qubit_index_x]

    ##  SECTION：----gamma门，作用于majorana_index----
    def u(self, majorana_index: int):
        """""
        input.majorana_index：目标费米子位
        influence：理想门，空操作。
        """""
        pass

    ##  SECTION：----gamma_prime门，作用于majorana_index----
    def v(self, majorana_index: int):
        """""
        input.majorana_index：目标费米子位
        influence：理想门，空操作。
        """""
        pass

    ##  SECTION：----i*gamma*gamma_prime门，作用于majorana_index----
    def n(self, majorana_index: int):
        """""
        input.majorana_index：目标费米子位
        influence：理想门，空操作。
        """""
        pass

    ##  SECTION：----P门，作用于majorana_index----
    def p(self, majorana_index: int):
        """""
        input.majorana_index：目标费米子位
        influence：交换费米子帧中对应位的X和Z分量。
        """""

        ##  PART：----数据预处理----
        assert isinteger(majorana_index) and 0<=majorana_index<self.majorana_number
        majorana_index_x=majorana_index*2
        majorana_index_z=majorana_index*2+1

        ##  PART：----P门作用----
        cache=self.frame[majorana_index_x].copy()
        self.frame[majorana_index_x]=self.frame[majorana_index_z]
        self.frame[majorana_index_z]=cache

    ##  SECTION：----CNOT门，作用于control_index,target_index，两者是qubits，前者是控制位----
    def cx(self, control_index, target_index):
        """""
        input.control_index, target_index：控制和目标量子位
        influence：更新Pauli帧：target_X += control_X, control_Z += target_Z。
        """""

        ##  PART：----数据预处理----
        assert isinteger(control_index) and 0<=control_index<self.pauli_number
        assert isinteger(target_index) and 0<=target_index<self.pauli_number
        control_qubit_index_x=self.majorana_number*2+control_index*2
        control_qubit_index_z=self.majorana_number*2+control_index*2+1
        target_qubit_index_x=self.majorana_number*2+target_index*2
        target_qubit_index_z=self.majorana_number*2+target_index*2+1

        ##  PART：----CNOT门作用----
        self.frame[target_qubit_index_x]+=self.frame[control_qubit_index_x]
        self.frame[control_qubit_index_z]+=self.frame[target_qubit_index_z]

    ##  SECTION：----CN-NOT门，作用于control_index,target_index，前者是fermionic site控制位，后者是qubit目标位----
    def cnx(self, control_index, target_index):
        """""
        input.control_index：费米子控制位
        input.target_index：量子位目标
        influence：根据混合门规则更新费米子-量子位帧。
        """""

        ##  PART：----数据预处理----
        assert isinteger(control_index) and 0<=control_index<self.majorana_number
        assert isinteger(target_index) and 0<=target_index<self.pauli_number
        control_majorana_index_x=control_index*2
        control_majorana_index_z=control_index*2+1
        target_qubit_index_x=self.majorana_number*2+target_index*2
        target_qubit_index_z=self.majorana_number*2+target_index*2+1

        ##  PART：----CN-NOT门作用----
        target_x= self.frame[target_qubit_index_x]+self.frame[control_majorana_index_x]+self.frame[control_majorana_index_z]
        control_x=self.frame[control_majorana_index_x]+self.frame[target_qubit_index_z]
        control_z=self.frame[control_majorana_index_z]+self.frame[target_qubit_index_z]
        self.frame[target_qubit_index_x]=target_x
        self.frame[control_majorana_index_x]=control_x
        self.frame[control_majorana_index_z]=control_z

    ##  SECTION：----CN-N门，作用于control_index,target_index，前者是fermionic site控制位，后者是fermionic site目标位----
    def cnn(self, control_index, target_index):
        """""
        input.control_index, target_index：控制和目标费米子位
        influence：根据费米子门规则更新帧。
        """""

        ##  PART：----数据预处理----
        assert isinteger(control_index) and 0<=control_index<self.majorana_number
        assert isinteger(target_index) and 0<=target_index<self.majorana_number
        control_majorana_index_x=control_index*2
        control_majorana_index_z=control_index*2+1
        target_majorana_index_x=target_index*2
        target_majorana_index_z=target_index*2+1

        ##  PART：----CN-N门作用----
        control_x= self.frame[control_majorana_index_x]+self.frame[target_majorana_index_x]+self.frame[target_majorana_index_z]
        control_z= self.frame[control_majorana_index_z]+self.frame[target_majorana_index_x]+self.frame[target_majorana_index_z]
        target_x= self.frame[target_majorana_index_x]+self.frame[control_majorana_index_x]+self.frame[control_majorana_index_z]
        target_z= self.frame[target_majorana_index_z]+self.frame[control_majorana_index_x]+self.frame[control_majorana_index_z]
        self.frame[control_majorana_index_x]=control_x
        self.frame[control_majorana_index_z]=control_z
        self.frame[target_majorana_index_x]=target_x
        self.frame[target_majorana_index_z]=target_z

    ##  SECTION：----Braid门，前者是fermionic site控制位，后者是fermionic site目标位----
    def braid(self, control_index, target_index, *args):
        """""
        input.control_index, target_index：费米子位
        influence：交换费米子帧中的特定分量。
        """""

        ##  PART：----数据预处理----
        assert isinteger(control_index) and 0<=control_index<self.majorana_number
        assert isinteger(target_index) and 0<=target_index<self.majorana_number
        control_majorana_index_z=control_index*2+1
        target_majorana_index_x=target_index*2

        ##  PART：----Braid门作用----
        cache=self.frame[control_majorana_index_z].copy()
        self.frame[control_majorana_index_z]=self.frame[target_majorana_index_x]
        self.frame[target_majorana_index_x]=cache

    ##  SECTION：----执行pauli_index上的X-error----
    def error(self, p,error_pattern:str,index_0:int,index_1:int=None):
        """""
        input.pauli_index_0,pauli_index_1：目标量子位
        input.string_0,string_1：目标量子位的pauli字符串
        input.p：错误概率
        influence：以概率p在帧上施加一个X错误。
        """""

        ##  PART：----数据预处理----
        pass

        ##  PART：----错误作用----
        if np.random.rand()<p:
            string_0=error_pattern[0]
            string_1=error_pattern[1]
            if string_0=='X':
                self.frame[index_0*2+self.majorana_number*2]+=self.GF(1)
            elif string_0=='Y':
                self.frame[index_0*2+self.majorana_number*2]+=self.GF(1)
                self.frame[index_0*2+self.majorana_number*2+1]+=self.GF(1)
            elif string_0=='Z':
                self.frame[index_0*2+self.majorana_number*2+1]+=self.GF(1)
            elif string_0=='U':
                self.frame[index_0*2]+=self.GF(1)
            elif string_0=='V':
                self.frame[index_0*2+1]+=self.GF(1)
            elif string_0=='N':
                self.frame[index_0*2]+=self.GF(1)
                self.frame[index_0*2+1]+=self.GF(1)
            else:
                raise ValueError(f"string_0={string_0} is not valid.")
            if string_1=='X':
                self.frame[index_1*2+self.majorana_number*2]+=self.GF(1)
            elif string_1=='Y':
                self.frame[index_1*2+self.majorana_number*2]+=self.GF(1)
                self.frame[index_1*2+self.majorana_number*2+1]+=self.GF(1)
            elif string_1=='Z':
                self.frame[index_1*2+self.majorana_number*2+1]+=self.GF(1)
            elif string_1=='U':
                self.frame[index_1*2]+=self.GF(1)
            elif string_1=='V':
                self.frame[index_1*2+1]+=self.GF(1)
            elif string_1=='N':
                self.frame[index_1*2]+=self.GF(1)
                self.frame[index_1*2+1]+=self.GF(1)
            elif string_1=='_':
                pass
            else:
                raise ValueError(f"string_0={string_0} or string_1={string_1} is not valid.")

    ##  SECTION：----重置0态----
    def reset(self, pauli_index):
        """""
        input.pauli_index：目标量子位
        influence：将帧中对应量子位分量清零，并随机引入一个Z错误（模拟测量后重置）。
        """""

        ##  PART：----数据预处理----
        assert isinteger(pauli_index) and 0<=pauli_index<self.pauli_number
        qubit_index_x=self.majorana_number*2+pauli_index*2
        qubit_index_z=self.majorana_number*2+pauli_index*2+1

        ##  PART：----重置0态----
        self.frame[qubit_index_x]=0
        self.frame[qubit_index_z]=0
        if np.random.rand()<0.5:
            self.frame[qubit_index_z]=self.GF(1)



import csv
import multiprocessing
import os
import time
from multiprocessing import Pool
import numpy as np
import pymatching
import stim
import stimbposd
from extendedstim.Circuit.Circuit import Circuit


#%%  APPENDIX：====尝试导入Tesseract类====
##  导入尝试
try:
    from tesseract_decoder import tesseract
    import tesseract_decoder

##  无法导入的情况下用BP-OSD替代
except ImportError:
    print("Tesseract decoder not found")


    class tesseract:

        def __init__(self):
            pass

        @staticmethod
        def TesseractConfig(dem, pqlimit, det_orders, det_beam=50, beam_climbing=False, no_revisit_dets=True):
            return Config(dem, det_beam)


    class Config:

        def __init__(self, dem: stim.DetectorErrorModel, det_beam):
            self.dem = dem
            self.det_beam = det_beam

        def compile_decoder(self):
            return stimbposd.bp_osd.BPOSD(model=self.dem, bp_method='min_sum', max_bp_iters=self.det_beam)

sample_number_now = multiprocessing.Value('i', 0)
error_number_now = multiprocessing.Value('i', 0)


def Simulation(circuit: Circuit, p_noise_list, p_measure_list, allowed_variant_ratios, decoder_name, decoder_params=None):
    """""        
    Tesseract decoder参数说明：
    {
    pqlimit：int对象，次数
    det_beam：int对象，分支查询数目
    beam_climbing：bool对象，是否梯度方法
    num_det_orders=num_det_orders，不知道什么阶数
    no_revisit_dets=no_revisit_dets，是否避免重复访问
    }
    ----------------------------------
    BP-OSD decoder参数说明：
    {
    bp_method：str对象，BP更新方法
    1. "product_sum"
    2. "min_sum"
    3. "min_sum_log"
    max_bp_iters：int对象，BP迭代次数，by default，50
    osd_method：str对象
        1. "osd_0"
        2. "osd_e": exhaustive OSD
        3. "osd_cs": combination-sweep OSD
        4. "osd_e"
    osd_order：int对象，OSD置换数目
    }
    """""

    ##  生成线路执行样本和每个样本的预测样本
    prototype = circuit.prototype
    group_number = os.cpu_count() // 2

    logical_error_rates = []
    variants = []
    for i in range(len(p_noise_list)):
        ##  计算解码事件并得到一次样本数量
        dem = prototype2dem(prototype, p_noise_list[i], p_measure_list[i])
        start_time = time.time()
        task(dem, decoder_name, 20, decoder_params=decoder_params, test=True)
        end_time = time.time()
        duration = (end_time - start_time) / 100
        print('解码一个样本的时间为：', duration, '秒')
        if 50 / duration > 500000:
            sample_in_one_group_number = 500000
        elif 50 / duration < 10:
            sample_in_one_group_number = 10
        else:
            sample_in_one_group_number = int(50 / duration)

        lock = multiprocessing.Manager().Lock()
        global sample_number_now
        global error_number_now
        sample_number_now.value = 0
        error_number_now.value = 0
        with Pool(processes=group_number) as pool:
            results = [pool.apply_async(task, args=(dem, decoder_name, sample_in_one_group_number, lock, allowed_variant_ratios[i], decoder_params)) for _
                       in range(group_number)]
            _ = [result.get() for result in results]
        error_rate = error_number_now.value / sample_number_now.value
        sigma = np.sqrt((error_number_now.value / sample_number_now.value - (
                error_number_now.value / sample_number_now.value) ** 2) / sample_number_now.value) * 2 * 3.2905
        logical_error_rates.append(error_rate)
        variants.append(sigma)

    return p_noise_list,p_measure_list,logical_error_rates, variants


def prototype2dem(prototype, p_noise, p_measure):
    ##  合并同类项
    dem_str = ''
    dem_temp = []
    for i, temp in enumerate(prototype):
        if temp[2] == 'DEP1':
            p_temp = (1 - np.sqrt(1 - 4 * p_noise / 3)) / 2
        elif temp[2] == 'DEP2':
            p_temp = (1 / 2) * (1 - (1 - 16 * p_noise / 15) ** (1 / 8))
        elif temp[2] == 'ONE_ERROR':
            p_temp = p_noise
        elif temp[2] == 'M_ERROR':
            p_temp = p_measure
        else:
            raise ValueError
        flag = True
        for j, item in enumerate(dem_temp):
            if temp[1] == item[1]:
                item[0] = p_temp * (1 - item[0]) + (1 - p_temp) * item[0]
                flag = False
                break
        if flag:
            dem_temp.append([p_temp, temp[1]])

    ##  生成检测错误模型
    for temp in dem_temp:
        dem_str += ('\n' + f'error({temp[0]})' + temp[1])
    return stim.DetectorErrorModel(dem_str)


# %%  CHAPTER：====并行计算错误率====
def task(dem, decoder_name, number, lock=None, error_bar=None, decoder_params=None, test=False):
    ##  生成线路执行样本和每个样本的预测样本
    sampler = dem.compile_sampler()  # 采样器

    if decoder_name == 'bposd':
        max_bp_iters = 50
        bp_method = 'min_sum'
        osd_method = 'osd_e'
        osd_order = 60
        if decoder_params is not None:
            if 'max_bp_iters' in decoder_params:
                max_bp_iters = decoder_params['max_bp_iters']
            if 'bp_method' in decoder_params:
                bp_method = decoder_params['bp_method']
            if 'osd_method' in decoder_params:
                osd_method = decoder_params['osd_method']
            if 'osd_order' in decoder_params:
                osd_order = decoder_params['osd_order']
        decoder = stimbposd.bp_osd.BPOSD(model=dem, bp_method=bp_method, max_bp_iters=max_bp_iters, osd_method=osd_method, osd_order=osd_order)

    elif decoder_name == 'tesseract':
        det_beam = 15
        beam_climbing = True
        no_revisit_dets = True
        num_det_orders = 16

        if decoder_params is not None:
            if 'det_beam' in decoder_params:
                det_beam = decoder_params['det_beam']
            if 'beam_climbing' in decoder_params:
                beam_climbing = decoder_params['beam_climbing']
            if 'num_det_orders' in decoder_params:
                num_det_orders = decoder_params['num_det_orders']
            if 'no_revisit_dets' in decoder_params:
                no_revisit_dets = decoder_params['no_revisit_dets']

        config = tesseract.TesseractConfig(dem=dem, pqlimit=200_000, det_beam=det_beam, beam_climbing=beam_climbing,
                                           det_orders=tesseract_decoder.utils.build_det_orders(
                                               dem=dem,
                                               num_det_orders=num_det_orders,
                                               method=tesseract_decoder.utils.DetOrder.DetIndex,
                                           ),
                                           no_revisit_dets=no_revisit_dets, )
        decoder = config.compile_decoder()

    elif decoder_name == 'matching':
        decoder = pymatching.Matching()
        decoder.from_detector_error_model(dem)
    else:
        raise NotImplementedError

    if test:
        detector_data, obs_data, error_data = sampler.sample(shots=number)  # 样本
        valid_detector_data = detector_data[
            np.where(np.logical_or(np.any(obs_data == True, axis=1), np.any(detector_data == True, axis=1)))[0]]  # 有效样本
        valid_obs_data = obs_data[np.where(np.logical_or(np.any(obs_data == True, axis=1), np.any(detector_data == True, axis=1)))[0]]  # 有效样本
        predictions = decoder.decode_batch(valid_detector_data)  # 解码器对每个样本的预测

        ##  计算逻辑错误率
        num_errors = 0
        for shot in range(len(valid_detector_data)):
            actual_for_shot = valid_obs_data[shot]
            predicted_for_shot = predictions[shot]
            if not np.array_equal(actual_for_shot, predicted_for_shot):
                num_errors += 1

    else:
        while True:
            global sample_number_now
            global error_number_now
            detector_data, obs_data, error_data = sampler.sample(shots=number)  # 样本
            valid_detector_data = detector_data[
                np.where(np.logical_or(np.any(obs_data == True, axis=1), np.any(detector_data == True, axis=1)))[0]]  # 有效样本
            valid_obs_data = obs_data[np.where(np.logical_or(np.any(obs_data == True, axis=1), np.any(detector_data == True, axis=1)))[0]]  # 有效样本
            predictions = decoder.decode_batch(valid_detector_data)  # 解码器对每个样本的预测

            ##  计算逻辑错误率
            num_errors = 0
            for shot in range(len(valid_detector_data)):
                actual_for_shot = valid_obs_data[shot]
                predicted_for_shot = predictions[shot]
                if not np.array_equal(actual_for_shot, predicted_for_shot):
                    num_errors += 1
            lock.acquire()
            sample_number_now.value += number
            error_number_now.value += num_errors
            error_rate = error_number_now.value / sample_number_now.value
            sigma = np.sqrt((error_number_now.value / sample_number_now.value - (
                    error_number_now.value / sample_number_now.value) ** 2) / sample_number_now.value)
            number_temp = sample_number_now.value
            lock.release()
            if sigma < error_bar * error_rate:
                break



"""""
模块作用：实现Majorana CSS码的距离估计、逻辑算符推导与若干构造器（由线性码/校验矩阵/标准例子）。
"""""
import numpy as np
from extendedstim.Code.PrimitiveCode.MajoranaCode import MajoranaCode
from extendedstim.Code.PrimitiveCode.QuantumCSSCode import QuantumCSSCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.tools.GaloisTools import orthogonalize, mip_distance_caculator, minus


class MajoranaCSSCode(MajoranaCode, QuantumCSSCode):

    # %%  CHAPTER：===构造方法===
    def __init__(self, generators_x, generators_z, physical_number):
        """""
        input.generators_x：X类Majorana稳定子
        input.generators_z：Z类Majorana稳定子
        input.physical_number：费米子位数
        """""
        QuantumCSSCode.__init__(self, generators_x, generators_z, physical_number)

    # %%  CHAPTER：===属性方法===
    ##  SECTION：----求码距----
    @property
    def distance(self):
        """""
        output：int，随机估计（或MIP）
        """""
        return mip_distance_caculator(self.check_matrix,minus(self.check_matrix.null_space(),self.check_matrix))

    ##  SECTION：----求码距（x方向）----
    @property
    def distance_x(self):
        """""
        output：int，基于MIP的精确/上界估计
        """""
        return mip_distance_caculator(self.check_matrix_x,minus(self.check_matrix_x.null_space(),self.check_matrix_x))

    ##  SECTION：----求码距（z方向）----
    @property
    def distance_z(self):
        """""
        output：int，基于MIP的精确/上界估计
        """""
        return mip_distance_caculator(self.check_matrix_z,minus(self.check_matrix_z.null_space(),self.check_matrix_z))

    ##  SECTION：----求逻辑算符----
    @property
    def logical_operators(self)->'list[MajoranaOperator]':
        """""
        output：np.array[MajoranaOperator]，X向和Z向配对拼接
        """""
        _=self._logical_operators_x
        return self._logical_operators_x+self._logical_operators_z

    ##  SECTION：----求逻辑算符（x方向）----
    @property
    def logical_operators_x(self)->'list[MajoranaOperator]':
        """""
        output：np.array[MajoranaOperator]，X向逻辑算符
        """""
        if self._logical_operators_x is None or len(self._logical_operators_x)==0:
            matrix = self.check_matrix_x
            codewords = matrix.null_space()
            independent_null_basis_list = []
            for vec in codewords:
                rank_before = np.linalg.matrix_rank(matrix)
                matrix = np.vstack([matrix, vec])
                if np.linalg.matrix_rank(matrix) == rank_before + 1:
                    independent_null_basis_list.append(vec)
            basis_list = orthogonalize(independent_null_basis_list)
            majorana_logical_operators_x = []
            majorana_logical_operators_z = []
            for i in range(len(basis_list)):
                occupy_temp=np.where(basis_list[i]==1)[0]
                temp = MajoranaOperator.HermitianOperatorFromOccupy(occupy_temp,[])
                majorana_logical_operators_x.append(temp)
                temp = MajoranaOperator.HermitianOperatorFromOccupy([],occupy_temp)
                majorana_logical_operators_z.append(temp)
            self._logical_operators_x = majorana_logical_operators_x
            self._logical_operators_z = majorana_logical_operators_z
            return self._logical_operators_x
        else:
            return self._logical_operators_x

    ##  SECTION：----求逻辑算符（z方向）----
    @property
    def logical_operators_z(self)->'list[MajoranaOperator]':
        """""
        output：np.array[MajoranaOperator]，Z向逻辑算符
        """""
        if self._logical_operators_z is None or len(self._logical_operators_z)==0:
            _=self.logical_operators_x
            return self._logical_operators_z
        else:
            return self._logical_operators_z

    #%%  CHAPTER：===静态方法===
    ##  SECTION：----从校验矩阵构造Majorana CSS code----
    @staticmethod
    def FromCheckMatrix(check_matrix):
        """""
        input.H：GF(2) (m,2n)
        output：MajoranaCSSCode
        """""
        generators_x = []
        generators_z = []
        for i in range(len(check_matrix)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromVector(check_matrix[i]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromVector(check_matrix[i]))
        physical_number=check_matrix.shape[1]
        return MajoranaCSSCode(generators_x, generators_z, physical_number)



import galois
import numpy as np
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.PauliCSSCode import PauliCSSCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator


class BicycleCode:
    def __init__(self, majorana_number:int, weight:int, logical_number:int, seed:int)->None:
        ##  SECTION：----数据预处理-----
        majorana_number=int(majorana_number)
        weight=int(weight)
        logical_number=int(logical_number)
        seed=int(seed)

        ##  SECTION：----生成校验矩阵-----
        np.random.seed(seed)
        H, diff_set = generate_matrix(majorana_number, weight, majorana_number-logical_number)
        self.seed=seed
        self.weight=weight
        self.H=H


#%%  CHAPTER：====构造Majorana bicycle code====
class MajoranaBicycleCode(MajoranaCSSCode,BicycleCode):
    """""
    self.majorana_number：int对象，fermionic sites的数目（偶数）
    self.weight：int对象，稳定子算符权重 (偶数)
    self.logical_number：int对象，逻辑位数目 (偶数)
    self.seed：int对象，随机种子
    """""

    __slots__ = ["majorana_number", "weight", "logical_number", "seed"]

    ##  SECTION：----构造方法-----
    def __init__(self, majorana_number:int, weight:int, logical_number:int, seed:int)->None:
        """""
        input.majorana_number：fermionic sites的数目（偶数）
        input.weight：稳定子算符权重 (偶数)
        input.logical_number：逻辑位数目 (偶数)
        input.seed：随机种子
        """""
        BicycleCode.__init__(self,majorana_number, weight, logical_number, seed)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromOccupy([],np.where(self.H[i] != 0)[0]))
        MajoranaCSSCode.__init__(self,generators_x,generators_z,majorana_number)

    def __str__(self):
        return f"MajoranaBicycleCode({self.majorana_number},{self.weight},{self.majorana_number-self.logical_number},{self.seed})"

class PauliBicycleCode(PauliCSSCode,BicycleCode):
    """""
    self.pauli_number：int对象，qubits的数目（偶数）
    self.weight：int对象，稳定子算符权重 (偶数)
    self.logical_number：int对象，逻辑位数目 (偶数)
    self.seed：int对象，随机种子
    """""

    __slots__ = ["pauli_number", "weight", "logical_number", "seed"]

    #%%  CHAPTER：====构造方法====
    def __init__(self, pauli_number:int, weight:int, logical_number:int, seed:int)->None:
        """""
        input.pauli_number：qubits的数目（偶数）
        input.weight：稳定子算符权重 (偶数)
        input.logical_number：逻辑位数目 (偶数)
        input.seed：随机种子
        """""

        BicycleCode.__init__(self,pauli_number, weight, logical_number, seed)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(PauliOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(PauliOperator.HermitianOperatorFromOccupy([],np.where(self.H[i] != 0)[0]))
        PauliCSSCode.__init__(self,generators_x,generators_z,pauli_number)


#%%   CHAPTER：====创建01循环矩阵=====
def create_circulant_matrix(difference_set, size):
    """""
    input.difference_set：差集（独特差值）
    input.size：矩阵大小 n
    output：numpy.ndarray (n,n) 01循环矩阵
    """""
    matrix = np.zeros((size, size), dtype=int)
    for i in range(size):
        for d in difference_set:
            j = (i + d) % size
            matrix[i, j] = 1
    return matrix


#%%   CHAPTER：====生成满足唯一差值特性的01循环矩阵====
def generate_matrix(physical_number, row_weight, row_number):
    """""
    input.N：总参数（偶数）
    input.k：行权目标（偶数）
    input.M：保留行数 (< N/2)
    output：(GF(2)矩阵, 差集)
    """""

    # 验证参数
    if physical_number % 2 != 0:
        raise ValueError("physical_number必须是偶数")
    if row_weight % 2 != 0:
        raise ValueError("row_weight必须是偶数")
    if row_number >= physical_number / 2:
        raise ValueError("row_number必须小于physical_number/2")

    n = physical_number // 2  # 循环矩阵大小
    k_half = row_weight // 2  # C的行权重

    # 步骤1: 生成满足唯一差值特性的差集
    difference_set = generate_difference_set(n, k_half)

    # 步骤2: 创建循环矩阵C和其转置
    C = create_circulant_matrix(difference_set, n)
    C_T = C.T

    # 步骤3: 构造H0 = [C, C_T]
    H0 = np.hstack((C, C_T))

    # 步骤4: 删除行以实现均匀列权重
    rows_to_keep = list(range(H0.shape[0]))

    # 计算初始列权重
    col_weights = np.sum(H0, axis=0)

    # 计算需要删除的行数
    rows_to_delete = H0.shape[0] - row_number

    # 贪心算法删除行，使列权重均匀
    for _ in range(rows_to_delete):
        best_row = -1
        best_variance = float('inf')

        # 尝试删除每一行，找到使列权重方差最小的行
        for row in rows_to_keep:
            # 临时删除该行
            temp_weights = col_weights - H0[row, :]
            variance = np.var(temp_weights)

            if variance < best_variance:
                best_variance = variance
                best_row = row

        # 删除最佳行
        rows_to_keep.remove(best_row)
        col_weights -= H0[best_row, :]

    # 创建最终矩阵
    H = H0[rows_to_keep, :]

    # 转换为GF(2)矩阵
    GF2 = galois.GF2
    H_gf2 = GF2(H)

    return H_gf2, difference_set


#%% CHAPTER：====生成满足唯一差值特性的差集====
def generate_difference_set(n, k):
    """""生成满足唯一差值特性的差集
    input.n：环大小
    input.k：差集目标大小
    output：排序后的差集列表
    """""

    # 初始化差集
    diff_set = [0]

    # 记录已使用的差值
    used_differences = set()

    while len(diff_set) < k:
        candidate = np.random.randint(1, n - 1)
        valid = True

        # 检查与现有元素的所有差值是否唯一
        for d in diff_set:
            diff1 = (candidate - d) % n
            diff2 = (d - candidate) % n

            if diff1 in used_differences or diff2 in used_differences:
                valid = False
                break

        if valid:
            # 添加新元素并记录差值
            diff_set.append(candidate)
            for d in diff_set[:-1]:
                diff1 = (candidate - d) % n
                diff2 = (d - candidate) % n
                used_differences.add(diff1)
                used_differences.add(diff2)

    return sorted(diff_set)


import numpy as np
import galois
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.PauliCSSCode import PauliCSSCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator


#%%  CHAPTER：====构造Double Unamed Code====
class DoubleReflectionCode:
    """""
    self.l：循环矩阵的大小
    self.m：循环矩阵的大小
    self.H：校验矩阵
    """""

    def __init__(self,l:int,m:int,x_list_a:list[int],y_list_a:list[int],x_list_b:list[int],y_list_b:list[int],p_list_x_a:list[int],p_list_y_a:list[int],p_list_x_b:list[int],p_list_y_b:list[int]) -> None:
        """""
        input.l：循环矩阵的大小
        input.m：循环矩阵的大小
        input.x_list_a：A中x项的移位幂次
        input.y_list_a：A中y项的移位幂次
        input.x_list_b：B中x项的移位幂次
        input.y_list_b：B中y项的移位幂次
        input.r_list_a：A中对应项的P矩阵幂次，可以是[r_x_list, r_y_list]或平铺列表
        input.r_list_b：B中对应项的P矩阵幂次，可以是[r_x_list, r_y_list]或平铺列表
        output：无
        influence：构造校验矩阵self.H
        """""
        
        ##  PART：----数据预处理----
        self.l=l
        self.m=m
        self.x_list_a=x_list_a
        self.y_list_a=y_list_a
        self.x_list_b=x_list_b
        self.y_list_b=y_list_b
        self.p_list_x_a=p_list_x_a
        self.p_list_y_a=p_list_y_a
        self.p_list_x_b=p_list_x_b
        self.p_list_y_b=p_list_y_b

        ##  PART：----构造矩阵A和B----
        A = self._construct_matrix(l, m, x_list_a, y_list_a, p_list_x_a, p_list_y_a)
        B = self._construct_matrix(l, m, x_list_b, y_list_b, p_list_x_b, p_list_y_b)
        
        ##  PART：----验证Double条件----
        if self._check_conditions(A, B):
            A_T = A.T
            B_T = B.T
            H_top = np.hstack((A, B, A_T, B_T))
            H_bottom = np.hstack((B_T, A_T, B, A))
            self.H = galois.GF(2)(np.vstack((H_top, H_bottom)))
        else:
            self.H = None

    ##  SECTION：----辅助函数：解析r_list----
    def _parse_r_list(self, r_list, len_x, len_y):
        if len(r_list) == 2 and isinstance(r_list[0], (list, tuple, np.ndarray)):
             return r_list[0], r_list[1]
        elif len(r_list) == len_x + len_y:
             return r_list[:len_x], r_list[len_x:]
        else:
             raise ValueError(f"r_list length {len(r_list)} does not match expected length {len_x}+{len_y} or structure [list, list]")

    ##  SECTION：----辅助函数：构造矩阵----
    def _construct_matrix(self, l, m, x_list, y_list, r_x, r_y):
        GF = galois.GF(2)
        # Term 1: sum x_i S_l^i P_l^a_ix \otimes I_m
        term1 = GF(np.zeros((l*m, l*m), dtype=int))
        for i, r in zip(x_list, r_x):
            S_l = shift(l, i)
            P_l = matrix_power_P(l, r)
            op = GF(np.kron(S_l @ P_l, np.eye(m, dtype=int)))
            term1 += op
            
        # Term 2: sum y_i I_l \otimes S_m^i P_m^a_iy
        term2 = GF(np.zeros((l*m, l*m), dtype=int))
        for i, r in zip(y_list, r_y):
            S_m = shift(m, i)
            P_m = matrix_power_P(m, r)
            op = GF(np.kron(np.eye(l, dtype=int), S_m @ P_m))
            term2 += op
            
        return term1 + term2

    ##  SECTION：----辅助函数：验证条件----
    def _check_conditions(self, A, B):
        # 1. A A^T = A^T A
        cond1 = np.array_equal(A @ A.T, A.T @ A)
        # 2. B B^T = B^T B
        cond2 = np.array_equal(B @ B.T, B.T @ B)
        # 3. A B = B A
        cond3 = np.array_equal(A @ B, B @ A)
            
        return cond1 and cond2 and cond3
        
    ##  SECTION：----字符串表示----
    def __str__(self):
        if self.H is None:
            return "DoubleUnamedCode(Construction Failed)"
        A_str='A='
        B_str='B='
        for i in range(len(self.x_list_a)):
            if i==0:
                A_str+=f'x^{self.x_list_a[i]}p^{self.p_list_x_a[i]}y^{self.y_list_a[i]}q^{self.p_list_y_a[i]}'
                B_str+=f'x^{self.x_list_b[i]}p^{self.p_list_x_b[i]}y^{self.y_list_b[i]}q^{self.p_list_y_b[i]}'
            else:
                A_str+=f'+x^{self.x_list_a[i]}p^{self.p_list_x_a[i]}y^{self.y_list_a[i]}q^{self.p_list_y_a[i]}'
                B_str+=f'+x^{self.x_list_b[i]}p^{self.p_list_x_b[i]}y^{self.y_list_b[i]}q^{self.p_list_y_b[i]}'
        return f"{self.l}|{self.m}|"+A_str+'|'+B_str


#%%  CHAPTER：====Majorana Double Unamed Code====
class MajoranaDoubleReflectionCode(MajoranaCSSCode, DoubleReflectionCode):
    
    ##  SECTION：----构造方法----
    def __init__(self,l:int,m:int,x_list_a:list[int],y_list_a:list[int],x_list_b:list[int],y_list_b:list[int],p_list_x_a:list[int],p_list_y_a:list[int],p_list_x_b:list[int],p_list_y_b:list[int]) -> None:
        DoubleReflectionCode.__init__(self, l, m, x_list_a, y_list_a, x_list_b, y_list_b, p_list_x_a, p_list_y_a, p_list_x_b, p_list_y_b)
        if self.H is None:
            raise ValueError("Code construction failed due to unsatisfied conditions.")
            
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromOccupy([],np.where(self.H[i]!=0)[0]))
        MajoranaCSSCode.__init__(self,generators_x,generators_z,self.l*self.m*4)


#%%  CHAPTER：====Pauli Double Unamed Code====
class PauliDoubleReflectionCode(PauliCSSCode, DoubleReflectionCode):
    
    ##  SECTION：----构造方法----
    def __init__(self,l:int,m:int,x_list_a:list[int],y_list_a:list[int],x_list_b:list[int],y_list_b:list[int],p_list_x_a:list[int],p_list_y_a:list[int],p_list_x_b:list[int],p_list_y_b:list[int]) -> None:
        DoubleReflectionCode.__init__(self, l, m, x_list_a, y_list_a, x_list_b, y_list_b, p_list_x_a, p_list_y_a, p_list_x_b, p_list_y_b)
        if self.H is None:
            raise ValueError("Code construction failed due to unsatisfied conditions.")
            
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(PauliOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(PauliOperator.HermitianOperatorFromOccupy([],np.where(self.H[i]!=0)[0]))
        PauliCSSCode.__init__(self,generators_x,generators_z,self.l*self.m*4)


# %%  CHAPTER：====用到的静态函数====
##  SECTION：----生成循环移位矩阵----
def shift(number:int, shift_amount:int):
    """""
    input.number：矩阵维度
    input.shift_amount：循环移位步长
    output：GF(2)矩阵，置换矩阵
    """""
    S = np.zeros((number, number), dtype=int)
    for i in range(number):
        S[i, (i + shift_amount) % number] = 1
    return galois.GF(2)(S)

##  SECTION：----生成P矩阵的幂----
def matrix_power_P(number:int, r:int):
    """""
    input.number：矩阵维度
    input.r：幂次
    output：GF(2)矩阵，P^r
    """""
    if r % 2 == 0:
        return galois.GF(2)(np.eye(number, dtype=int))
    else:
        # P is anti-diagonal
        return galois.GF(2)(np.fliplr(np.eye(number, dtype=int)))



import numpy as np
from galois import GF2
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.PauliCSSCode import PauliCSSCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator


class ColorCode:

    #%%  CHAPTER：----生成Majorana Color Code----
    def __init__(self, d: int) -> None:
        self.edge_stabilizers=[]
        self.d=d
        """""
        input.d：奇数，表示三角形网格的边长（distance）
        output：PauliCSSCode
        """""

        ##  基本数值罗列
        x_0=0
        y_0=0
        cell_vector=[np.array([0, 0], dtype=float), np.array([2, 0], dtype=float), np.array([0.5, np.sqrt(3)/2], dtype=float),
                     np.array([1.5, np.sqrt(3)/2], dtype=float), np.array([0.5, -np.sqrt(3)/2], dtype=float),
                     np.array([1.5, -np.sqrt(3)/2], dtype=float)]
        cells=[]
        x_now=0
        y_now=0

        ##  生成所有的六角晶格
        for i in range(d//2):
            for j in range(3):
                for k in range(d//2-i):
                    x_now_temp=x_now+k*3
                    cells.append([np.array([x_now_temp, y_now], dtype=float)+cell_vector[k] for k in range(6)])
                if j==0:
                    y_now+=np.sqrt(3)/2
                    x_now+=1.5
                elif j==1:
                    y_now+=np.sqrt(3)/2
                    x_now-=1.5
                elif j==2:
                    y_now+=np.sqrt(3)/2
                    x_now+=1.5

        ##  筛除不在三角形内的晶格
        cells_in_triangle=[]
        for i in range(len(cells)):
            cells_in_triangle.append([])
            for j in range(6):
                if judge_in_triangle(cells[i][j][0], cells[i][j][1],x_0,y_0,d):
                    cells_in_triangle[i].append(cells[i][j])
        position_list=[]

        stabilizer_list=[]
        ##  查找底层sites
        for i in range(len(cells_in_triangle)):
            for j in range(len(cells_in_triangle[i])):
                if abs(cells_in_triangle[i][j][1])<0.01:
                    flag=True
                    for k in range(len(position_list)):
                        if equal(cells_in_triangle[i][j], position_list[k]):
                            flag=False
                            break
                    if flag:
                        position_list.append(cells_in_triangle[i][j])

        ##  根据晶体生成stabilizers
        for i in range(len(cells_in_triangle)):
            stabilizer_temp=[]
            for j in range(len(cells_in_triangle[i])):
                flag=True
                index=0
                for k in range(len(position_list)):
                    if equal(cells_in_triangle[i][j], position_list[k]):
                        flag=False
                        index=k
                        break
                if flag:
                    stabilizer_temp.append(len(position_list))
                    position_list.append(cells_in_triangle[i][j])
                else:
                    stabilizer_temp.append(index)
            stabilizer_list.append(stabilizer_temp)

        self.H=GF2(np.zeros((len(stabilizer_list),len(stabilizer_list)),dtype=int))
        for i in range(len(stabilizer_list)):
            self.H[i][tuple(stabilizer_list[i])]=1


# %%  CHAPTER：====Majorana color code====
class MajoranaColorCode(MajoranaCSSCode,ColorCode):

    ##  SECTION：----构造方法----
    def __init__(self,d:int):
        ColorCode.__init__(self,d)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromOccupy([],np.where(self.H[i]!=0)[0]))
        MajoranaCSSCode.__init__(self,generators_x,generators_z,self.H.shape[1])


# %%  CHAPTER：====Pauli color code====
class PauliColorCode(PauliCSSCode,ColorCode):

    ##  SECTION：----构造方法----
    def __init__(self,d:int):
        ColorCode.__init__(self,d)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(PauliOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(PauliOperator.HermitianOperatorFromOccupy([],np.where(self.H[i]!=0)[0]))
        PauliCSSCode.__init__(self,generators_x,generators_z,self.H.shape[1])


def judge_in_triangle(x, y,x_0,y_0,d):
    """""
    input.x：float，x坐标
    input.y：float，y坐标
    output：bool，是否在三角形内
    """""
    x_1=x_0+d//2+d-1
    x_2=x_1/2
    y_2=np.sqrt(3)*x_2

    k_0=0
    b_0=0
    k_1=(y_2-y_0)/(x_2-x_0)
    b_1=0
    k_2=-k_1
    b_2=2*y_2
    if k_0*x+b_0<y+0.01 and k_1*x+b_1>y-0.01 and k_2*x+b_2>y-0.01:
        return True
    else:
        return False


def equal(a, b):
    """""
    input.a：float，a坐标
    input.b：float，b坐标
    output：bool，是否相等
    """""
    if abs(a[0]-b[0])+abs(a[1]-b[1])<0.01:
        return True
    else:
        return False



"""""
模块作用：实现构造
"""""
import galois
import numpy as np
from extendedstim.Code.PrimitiveCode.MajoranaCSSCode import MajoranaCSSCode
from extendedstim.Code.PrimitiveCode.PauliCSSCode import PauliCSSCode
from extendedstim.Physics.MajoranaOperator import MajoranaOperator
from extendedstim.Physics.PauliOperator import PauliOperator


# %%  CHAPTER：====构造校验矩阵的类====
class DoubleBicycleCode:

    ##  SECTION：----构造方法----
    def __init__(self,n:int,a_occupy:list,b_occupy:list)->None:
        """""
        input.n：循环矩阵的大小
        input.a_occupy：循环矩阵A中1的位置
        input.b_occupy：循环矩阵B中1的位置
        output：无（构造对象）
        """""

        ##  PART：----数据预处理----
        self.n=n
        self.a_occupy=a_occupy.copy()
        self.b_occupy=b_occupy.copy()
        self.a_occupy.sort()
        self.b_occupy.sort()

        ##  PART：----构造校验矩阵----
        a_list = [1 if i in self.a_occupy else 0 for i in range(n)]
        b_list = [1 if i in self.b_occupy else 0 for i in range(n)]
        H_up = np.hstack([cyclic_matrix(n, a_list), cyclic_matrix(n, b_list), cyclic_matrix(n, a_list).T, cyclic_matrix(n, b_list).T])
        H_down = np.hstack([cyclic_matrix(n, b_list).T, cyclic_matrix(n, a_list).T, cyclic_matrix(n, b_list), cyclic_matrix(n, a_list)])
        H = np.vstack([H_up, H_down])
        self.H=H

    ##  SECTION：----打印代码----
    def __str__(self):
        n = self.n

        S_str_a = 'A='
        for i in self.a_occupy:
            if i == 0:
                S_str_a += f"I_{{{n}}}"
            elif i == 1:
                S_str_a += f"S_{{{n}}}"
            else:
                S_str_a += f"S^{{{i}}}_{{{n}}}"

        S_str_b = 'B='
        for i in self.b_occupy:
            if i == 0:
                S_str_b += f"I_{{{n}}}"
            elif i == 1:
                S_str_b += f"S_{{{n}}}"
            else:
                S_str_b += f"S^{{{i}}}_{{{n}}}"

        return '$' + S_str_a + '$' + '\n' + '$' + S_str_b + '$'


# %%  CHAPTER：====Majorana double bicycle code====
class MajoranaDoubleBicycleCode(MajoranaCSSCode,DoubleBicycleCode):

    ##  SECTION：----构造方法----
    def __init__(self, n: int, a_occupy: list, b_occupy: list) -> None:
        DoubleBicycleCode.__init__(self,n,a_occupy,b_occupy)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromOccupy([],np.where(self.H[i] != 0)[0]))
        MajoranaCSSCode.__init__(self,generators_x,generators_z,self.n*4)


# %%  CHAPTER：====Pauli double bicycle code====
class PauliDoubleBicycleCode(PauliCSSCode,DoubleBicycleCode):

    ##  SECTION：----构造方法----
    def __init__(self, n: int, a_occupy: list, b_occupy: list) -> None:
        DoubleBicycleCode.__init__(self,n,a_occupy,b_occupy)
        generators_x=[]
        generators_z=[]
        for i in range(len(self.H)):
            generators_x.append(PauliOperator.HermitianOperatorFromOccupy(np.where(self.H[i]!=0)[0],[]))
            generators_z.append(PauliOperator.HermitianOperatorFromOccupy([],np.where(self.H[i] != 0)[0]))
        PauliCSSCode.__init__(self,generators_x,generators_z,self.n*4)


# %%  CHAPTER：====用到的静态函数====
##  SECTION：----生成循环移位矩阵----
def shift(number:int, shift:int):
    """""
    input.number：矩阵维度
    input.shift：循环移位步长
    output：GF(2)矩阵，置换矩阵
    """""

    S = np.zeros((number, number), dtype=int)
    for i in range(number):
        S[i, (i + shift) % number] = 1
    return galois.GF(2)(S)


##  SECTION：----创建GF(2)循环矩阵----
def cyclic_matrix(n, a_list):
    """""
    input.n：尺寸
    input.a_list：长度为n的01序列，指示多项式系数
    output：GF(2)循环矩阵 C(n,a)
    """""
    assert len(a_list) == n
    return galois.GF2(np.sum([a*(np.linalg.matrix_power(shift(n,1),i)) for i,a in enumerate(a_list)],axis=0))